/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode;

import teajay.bytecode.exceptions.ByteCodeException;

/**
 * The JVM base types.
 * The string value of these enum constants can be used in field descriptors to
 * represent their respective value.
 * @author Michael Kußmann <michi@outx.de>
 */
public enum BaseType {
    Void("V"),
    Boolean("Z"),
    Byte("B"),
    Short("S"),
    Char("C"),
    Int("I"),
    Float("F"),
    Double("D"),
    Long("J"),
    /**
     * Syntax: L&lt;fully qualified name&gt;;
     */
    Reference("L"),
    /**
     * Syntax: [...[ &lt; field descriptor &gt;
     */
    Array("[");
    
    private final String value;

    private BaseType(String value) {
        this.value = value;
    }
    
    public String getValue() {
        return value;
    }
    
    public char getChar() {
        return value.charAt(0);
    }
    
    public static BaseType parse(char c) throws ByteCodeException {
        switch (c) {
            case 'V':
                return Void;
            case 'B':
                return Byte;
            case 'C':
                return Char;
            case 'D':
                return Double;
            case 'F':
                return Float;
            case 'I':
                return Int;
            case 'J':
                return Long;
            case 'L':
                return Reference;
            case 'S':
                return Short;
            case 'Z':
                return Boolean;
            case '[':
                return Array;
            default:
                throw new ByteCodeException("unknown base type " + c);
        }
    }
    
    public static BaseType parseNoThrow(char c) {
        try {
            return parse(c);
        } catch (ByteCodeException ex) {
            return null;
        }
    }
}
