/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode.attributes.code;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import teajay.bytecode.OpCode;
import teajay.bytecode.Operand;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.exceptions.ByteCodeException;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
abstract class SimpleInstruction extends JVMInstruction {

    private OpCode opcode = null;
    private List<Operand> operands = Collections.emptyList();

    protected SimpleInstruction() {
    }

    protected final void setOpcode(OpCode opcode) {
        this.opcode = opcode;
    }

    protected final void setOperands(Operand... operands) {
        if (opcode != null) {
            opcode.validate(operands);
            this.operands = Arrays.asList(operands);
        } else {
            throw new RuntimeException("Cannot set operands when opcode hasn't yet been set.");
        }
    }

    @Override
    public final OpCode opcode() throws ByteCodeException {
        return opcode;
    }

    @Override
    protected final List<Operand> operands() {
        return operands;
    }

    @Override
    int getLength() throws ByteCodeException {
        int len = 1; //1 for the opcode
        for (Operand operand : operands()) {
            len += operand.getDefaultLengthAsOperand().size;
        }
        return len;
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        if (opcode != null) {
            opcode.write(out, cpool);

            for (Operand operand : operands()) {
                operand.writeIndex(out, cpool);
            }
        } else {
            throw new ByteCodeException("Trying to write the instruction but the opcode hasn't yet been set.");
        }
    }

    @Override
    boolean validate(ConstantPool cpool) throws ByteCodeException {
        if (opcode == null) {
            throw new ByteCodeException("The opcode hasn't yet been set: The validation has failed permanently.");
        }
        return true;
    }

    @Override
    int getPush() throws ByteCodeException {
        if (opcode != null) {
            if (opcode == OpCode.wide1 || opcode == OpCode.wide2) {
                if (operands().size() > 0) {
                    Operand first = operands().get(0);
                    if (first instanceof OpCode) {
                        return ((OpCode) first).num_push;
                    }
                }
                throw new RuntimeException("Invalid use of the wide instruction.");
            }
            if (opcode.num_push < 0) {
                throw new RuntimeException("Invalid use of SimpleInstruction: getPush()=" + opcode.num_push);
            }
            return opcode.num_push;
        } else {
            throw new ByteCodeException("The opcode hasn't yet been set: getPush() failed permanently");
        }
    }

    @Override
    int getPop() throws ByteCodeException {
        if (opcode != null) {
            if (opcode == OpCode.wide1 || opcode == OpCode.wide2) {
                if (operands().size() > 0) {
                    Operand first = operands().get(0);
                    if (first instanceof OpCode) {
                        return ((OpCode) first).num_pop;
                    }
                }
                throw new RuntimeException("Invalid use of the wide instruction.");
            }
            if (opcode.num_pop < 0) {
                throw new RuntimeException("Invalid use of SimpleInstruction: getPop()=" + opcode.num_pop);
            }
            return opcode.num_pop;
        } else {
            throw new ByteCodeException("The opcode hasn't yet been set: getPop() failed permanently");
        }
    }
}
