/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode.attributes.code;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.List;
import teajay.bytecode.ByteCodeWritable;
import teajay.bytecode.U1;
import teajay.bytecode.U2;
import teajay.bytecode.U4;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.ConstantPoolAddable;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.exceptions.FrameOffsetDeltaOutOfBounds;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public abstract class StackMapFrame implements ByteCodeWritable, ConstantPoolAddable {
   
    private JVMInstruction start;
    private StackMapFrame predecessor;
    private int offsetDelta = -1;
    private boolean added = false;

    protected StackMapFrame(JVMInstruction start) {
        this.start = start;
    }

    abstract U1 getFrameType() throws ByteCodeException;

    @Override
    public abstract void add(ConstantPool cpool) throws ByteCodeException;

    abstract U4 getLength() throws ByteCodeException;

    final U2 offsetDelta() throws ByteCodeException {
        if (offsetDelta < 0) {
            throw new ByteCodeException("Trying to access the offset delta, but it is not yet set.");
        }
        return new U2(offsetDelta);
    }

    final JVMInstruction getStart() {
        return start;
    }
    
    final void setStart(JVMInstruction start) {
        this.start = start;
    }
    
    final void setAdded(boolean added) {
        this.added = added;
    }
    
    final boolean isAdded() {
        return added;
    }
    
    void setPredecessor(StackMapFrame predecessor) {
        this.predecessor = predecessor;
    }
    
    StackMapFrame getPredecessor() throws ByteCodeException {
        if (predecessor == null) {
            throw new ByteCodeException("Trying to access the predecessor of a stack map frame, but it is not yet set!");
        }
        return this.predecessor;
    }

    abstract List<VerificationType> getStack() throws ByteCodeException;

    abstract List<VerificationType> getLocals() throws ByteCodeException;

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        getFrameType().write(out, cpool);
    }
    
    void validate() throws ByteCodeException {
    }

    final void setOffsetDelta(int delta) throws FrameOffsetDeltaOutOfBounds {
        if (delta < 0 || delta > 0xFFFF) {
            throw new FrameOffsetDeltaOutOfBounds("offset delta out of bounds: " + delta);
        }
        this.offsetDelta = delta;
    }

    final int getOffsetDelta() {
        return offsetDelta;
    }
}
