/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode.cpool;

import java.io.DataOutputStream;
import java.io.IOException;
import teajay.bytecode.ByteCodeWritable;
import teajay.bytecode.OpCode.OpSize;
import teajay.bytecode.Operand;
import teajay.bytecode.exceptions.ByteCodeException;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public abstract class CPoolEntry implements ByteCodeWritable, Operand, ConstantPoolAddable {

    private final CPoolTag tag;

    protected CPoolEntry(CPoolTag tag) {
        this.tag = tag;
    }
    
    /**
     * Returns this entry's constant pool tag.
     * @return this entry's constant pool tag.
     */
    public final CPoolTag getTag() {
        return tag;
    }

    public final U2Index getCIndex(ConstantPool cpool) throws ByteCodeException {
        U2Index index = cpool.find(this);
        if (index != null) {
            return index;
        } else {
            throw new ByteCodeException("no such constant pool entry");
        }
    }

    @Override
    public final long getAddress(ConstantPool cpool) throws ByteCodeException {
        return getCIndex(cpool).getValue();
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        out.writeByte(tag.intValue);
    }

    @Override
    public void writeIndex(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        getCIndex(cpool).write(out, cpool);
    }

    @Override
    public OpSize getDefaultLengthAsOperand() {
        return OpSize.U2;
    }

    @Override
    public boolean equals(Object obj) {
        if (obj instanceof CPoolEntry) {
            return ((CPoolEntry) obj).getTag() == getTag();
        } else {
            return false;
        }
    }
    
    @Override
    public int hashCode() {
        int hash = 7;
        hash = 37 * hash + (this.tag != null ? this.tag.hashCode() : 0);
        return hash;
    }

    @Override
    public String toString() {
        return this.getClass().getSimpleName() + ": tag=" + getTag();
    }

    /**
     * Adds itself to the constant pool <tt> pool </tt>.
     * @param pool
     * @throws teajay.bytecode.exceptions.ByteCodeException
     */
    @Override
    public void add(ConstantPool pool) throws ByteCodeException {
        pool.add(this);
    }
}
