/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode.cpool;

import static teajay.bytecode.BaseType.*;
import teajay.bytecode.exceptions.MalformedClassNameException;
import static teajay.bytecode.util.Tools.checkBinaryName;

/**
 * class_info constant pool entry. The class name has to be supplied in internal
 * jvm format. Example: java.lang.String becomes java/lang/String The JVM 7
 * spec. clearly states that all identifiers between the / must be unqualified
 * names. This would make it impossible to create class_infos for array types.
 * Thus this implementation only enforces class_infos not to contain any '.'s to
 * prevent confusion.
 *
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class ClassInfo extends U1U2Entry<Utf8Info> {

    /**
     * Creates a new ClassInfo instance.
     *
     * @param name internal class name
     * @throws MalformedClassNameException if the class name contains illegal
     * characters
     */
    public ClassInfo(Utf8Info name) throws MalformedClassNameException {
        super(CPoolTag.Class, name);
        if (!checkBinaryName(name.getValue())) {
            throw new MalformedClassNameException("ClassInfo expects a binary class or interface name: " + name.getValue());
        }
    }

    /**
     * @see #ClassInfo(Utf8Info)
     * @param name
     * @throws MalformedClassNameException
     */
    public ClassInfo(String name) throws MalformedClassNameException {
        this(new Utf8Info(name));
    }

    public static ClassInfo fromClass(Class<?> cl) throws MalformedClassNameException {
        return new ClassInfo(cl.getName().replace('.', '/'));
    }

    public Class<?> getAsClass() throws ClassNotFoundException {
        return getAsClass(Thread.currentThread().getContextClassLoader());
    }

    public Class<?> getAsClass(ClassLoader loader) throws ClassNotFoundException {
        if (isPrimitive()) {
            switch (this.getClassName()) {
                case "void":
                    return void.class;
                case "boolean":
                    return boolean.class;
                case "byte":
                    return byte.class;
                case "short":
                    return short.class;
                case "int":
                    return int.class;
                case "long":
                    return long.class;
                case "float":
                    return float.class;
                case "double":
                    return double.class;
                case "char":
                    return char.class;
                default:
                    throw new RuntimeException("unreachable statement reached");
            }
        } else {
            Class<?> cl = loader.loadClass(this.getClassName().replace('/', '.'));
            return cl;
        }
    }

    public String getClassName() {
        Utf8Info value = this.getValue();
        if (value == null) {
            return null;
        } else {
            return value.getValue();
        }
    }

    public String getSimpleName() {
        Utf8Info value = this.getValue();
        if (value == null) {
            return null;
        } else {
            String str = value.getValue();
            int index = str.lastIndexOf("/");
            if (index <= -1) {
                return str;
            } else {
                return str.substring(index + 1);
            }
        }
    }

    @Override
    public boolean equals(Object obj) {
        return (obj instanceof ClassInfo) && super.equals(obj);
    }

    @Override
    public int hashCode() {
        int hash = 5 * super.hashCode();
        return hash;
    }

    public boolean isArray() {
        return getClassName().startsWith("[");
    }

    public boolean isPrimitive() {
        switch (this.getClassName()) {
            case "void":
            case "boolean":
            case "byte":
            case "short":
            case "char":
            case "int":
            case "long":
            case "float":
            case "double":
                return true;
            default:
                return false;
        }
    }

    public String getPrimitiveChar() {
        switch (this.getClassName()) {
            case "void":
                return Void.getValue();
            case "boolean":
                return Boolean.getValue();
            case "byte":
                return Byte.getValue();
            case "short":
                return Short.getValue();
            case "char":
                return Char.getValue();
            case "int":
                return Int.getValue();
            case "long":
                return Long.getValue();
            case "float":
                return Float.getValue();
            case "double":
                return Double.getValue();
            default:
                return Reference.getValue();
        }
    }
}
