/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode.cpool;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Objects;
import teajay.bytecode.U2;
import teajay.bytecode.attributes.bootstrap.BootstrapMethodEntry;
import teajay.bytecode.exceptions.ByteCodeException;

/**
 * Implements the constant pool entry that serves as argument for a invokedynamic instruction.
 * This class is not cloneable!
 * @author Michael Kußmann <michi@outx.de>
 */
public class InvokeDynamicInfo extends CPoolEntry {
    private final U2 bootstrapIndex;
    private final NameAndTypeInfo<MethodDescriptor> nameAndType;
    
    public InvokeDynamicInfo(BootstrapMethodEntry bootstrapMethod, NameAndTypeInfo<MethodDescriptor> nameAndType) throws ByteCodeException {
        super(CPoolTag.InvokeDynamic);
        this.bootstrapIndex = bootstrapMethod.getIndex();
        this.nameAndType = nameAndType;
    }
    
    InvokeDynamicInfo(U2 bootstrapIndex, NameAndTypeInfo<MethodDescriptor> nameAndType) throws ByteCodeException {
        super(CPoolTag.InvokeDynamic);
        this.bootstrapIndex = bootstrapIndex;
        this.nameAndType = nameAndType;
    }
    
    public final U2 getBootstrapIndex() {
        return this.bootstrapIndex;
    }
    
    public final NameAndTypeInfo<MethodDescriptor> getNameAndType() {
        return this.nameAndType;
    }
    
    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        super.write(out, cpool);
        bootstrapIndex.write(out, cpool);
        nameAndType.getCIndex(cpool).write(out, cpool);
    }
    
    @Override
    public String toString() {
        return this.getClass().getSimpleName() + " (bootstrap_index = " + bootstrapIndex + ", name_and_type = " + nameAndType + ")";
    }
    
    @Override
    public boolean equals(Object obj) {
        if ((obj instanceof InvokeDynamicInfo) && super.equals(obj)) {
            InvokeDynamicInfo other = (InvokeDynamicInfo) obj;
            return other.bootstrapIndex.equals(bootstrapIndex) && other.getNameAndType().equals(getNameAndType());
        } else {
            return false;
        }
    }

    @Override
    public int hashCode() {
        int hash = 51 * super.hashCode();
        hash = 89 * hash + Objects.hashCode(this.bootstrapIndex);
        hash = 89 * hash + Objects.hashCode(this.nameAndType);
        return hash;
    }
    
    @Override
    public void add(ConstantPool pool) throws ByteCodeException {
        super.add(pool);
        nameAndType.add(pool);
    }
}
