/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode.cpool;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Objects;
import teajay.bytecode.ByteCodeWritable;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.exceptions.UTFOutOfBoundsException;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 * @param <T>
 */
public final class NameAndTypeInfo<T extends Utf8Info> extends CPoolEntry implements ByteCodeWritable {

    private final Utf8Info name;
    private final T descriptor;
    
    public NameAndTypeInfo(String name, T descriptor) throws UTFOutOfBoundsException {
        this(new Utf8Info(name), descriptor);
    }

    public NameAndTypeInfo(Utf8Info name, T descriptor) {
        super(CPoolTag.NameAndType);
        this.name = name;
        this.descriptor = descriptor;
    }

    public final Utf8Info getName() {
        return this.name;
    }

    public final T getDescriptor() {
        return this.descriptor;
    }
    
    public final boolean isMethodDescriptor() {
        return descriptor instanceof MethodDescriptor;
    }
    
    public final boolean isFieldDescriptor() {
        return descriptor instanceof FieldDescriptor;
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        super.write(out, cpool);
        name.getCIndex(cpool).write(out, cpool);
        descriptor.getCIndex(cpool).write(out, cpool);
    }

    @Override
    public String toString() {
        return this.getClass().getSimpleName() + " (name = " + name + ", descriptor = " + descriptor + ")";
    }

    @Override
    public boolean equals(Object obj) {
        if ((obj instanceof NameAndTypeInfo) && super.equals(obj)) {
            NameAndTypeInfo<?> other = (NameAndTypeInfo<?>) obj;
            return other.getName().equals(getName()) && other.getDescriptor().equals(getDescriptor());
        } else {
            return false;
        }
    }

    @Override
    public int hashCode() {
        int hash = 29 * super.hashCode();
        hash = 59 * hash + Objects.hashCode(this.name);
        hash = 59 * hash + Objects.hashCode(this.descriptor);
        return hash;
    }

    @Override
    public void add(ConstantPool pool) throws ByteCodeException {
        super.add(pool);
        name.add(pool);
        descriptor.add(pool);
    }
}
