/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package codegeneration.instruction;

import abstractsyntaxtree.InfixOpNode;
import abstractsyntaxtree.SyntaxNode;
import codegeneration.javavisitor.VariableInitializerVisitor;
import codegeneration.environment.ExpressionEnvironment;
import codegeneration.environment.Stage3Environment;
import datastructs.Pair;
import exceptions.CompilerException;
import teajay.runtime.exceptions.ResolveException;
import java.io.IOException;
import java.util.Arrays;
import java.util.Deque;
import java.util.logging.Level;
import java.util.logging.Logger;
import symbol.LocalEntry;
import teajay.bytecode.ClassFile;
import teajay.bytecode.attributes.code.ALoad;
import teajay.bytecode.attributes.code.GetStatic;
import teajay.bytecode.attributes.code.IConst;
import teajay.bytecode.attributes.code.IXOr;
import teajay.bytecode.attributes.code.InvokeStatic;
import teajay.bytecode.attributes.code.InvokeVirtual;
import teajay.bytecode.attributes.code.JVMInstructionList;
import teajay.bytecode.attributes.code.Swap;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.cpool.MethodDescriptor;
import teajay.bytecode.cpool.MethodRefInfo;
import teajay.bytecode.cpool.NameAndTypeInfo;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.fields.FieldInfo;
import teajay.bytecode.methods.MethodInfo;
import teajay.generics.check.TypeUtils;
import teajay.generics.parser.Type;
import teajay.methods.MethodFinder;
import teajay.util.io.ClassFileRef;
import teajay.util.io.ClassFinder;

/**
 *
 * @author Michael Burzan , Michael Kußmann
 */
public class InstructionCreator {

    /**
     * @deprecated 
     * @param stage3Env
     * @param type
     * @param paramaters
     * @param node
     * @return
     * @throws CompilerException 
     */
    public static Pair<MethodInfo, MethodRefInfo> CreateConstructor(Stage3Environment stage3Env, Type type, FieldDescriptor[] paramaters, SyntaxNode node) throws CompilerException {
        //TODO: Call 
        MethodRefInfo ref = null;
        MethodInfo cInfo = null;
        try {

            if (type instanceof Type.ClassType) {

                Type.ClassType cType = (Type.ClassType) type;

                ClassFile cFile = ClassFinder.instance().find(cType.name).get();

                if (cFile.isAbstract()) {
                    throw new CompilerException("It's not possible to create object of abstract type", node.getSymbol());
                }
                Pair<MethodInfo, MethodFinder.DistanceTailPair> findMethod = MethodFinder.findMethod("<init>", Arrays.asList(paramaters), cType.name,
                        stage3Env.getClassFile().getThis().getClassName(), null, MethodFinder.QueryType.NON_VIRTUAL);
                if (findMethod != null) {
                    cInfo = findMethod.value1;
                }
                if (cInfo == null) {
                    throw new CompilerException("Constructor missmatch for class " + cType.name + " " + Arrays.toString(paramaters), node.getSymbol());
                }
                ref = cInfo.makeMethodRef();
            } else {
                throw new CompilerException("It's not possible to create object of class " + type, node.getSymbol());
            }
        } catch (ByteCodeException | IOException | ResolveException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
        return new Pair<>(cInfo, ref);
    }

    @Deprecated
    public static MethodRefInfo CreateConstructor(String typeName, String... parameters) throws CompilerException {
        FieldDescriptor[] field = new FieldDescriptor[parameters.length];
        int i = 0;
        MethodDescriptor desc;
        if (parameters[0] != null) {
            desc = new MethodDescriptor(FieldDescriptor.getVoid(),
                    field);
            for (String p : parameters) {
                field[i] = FieldDescriptor.fromString(p);
                i++;
            }
        } else {
            desc = new MethodDescriptor(FieldDescriptor.getVoid());
        }


        NameAndTypeInfo nameInfo = new NameAndTypeInfo("<init>", desc);
        MethodRefInfo info = null;
        try {
            info = new MethodRefInfo(typeName, nameInfo);
        } catch (ByteCodeException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
        return info;
    }
    /**
     * debug method 
     * @param info
     * @param varName
     * @throws CompilerException 
     */
    public static void MakeDebugOutputOfLocalReference(MethodInfo info, String varName) throws CompilerException {
        ClassFileRef ref = ClassFinder.instance().find("java/lang/System");
        if (ref != null) {
            ClassFile file;
            try {
                file = ref.get();
                FieldInfo finfo = file.getField("out");
                ClassInfo cinfo = finfo.getDescriptor().getClassInfo();
                LocalEntry entry = null;// SymbolFacade.instance().getLocalTable().getEntry(varName);
                MethodInfo printlnMethodInfo = symbol.MethodResolver.existMethod(null,
                        cinfo.getClassName(), "println", entry.getQID());

                NameAndTypeInfo nameAndType = new NameAndTypeInfo("println", printlnMethodInfo.getDescriptor());
                MethodRefInfo mRef = new MethodRefInfo(cinfo, nameAndType);

                JVMInstructionList instructionList = info.getCode().getInstructionList();
                instructionList.add(new GetStatic(finfo.getFieldRefInfo(file.getThis())));
                instructionList.add(new ALoad(entry.getIndex()));
                instructionList.add(new InvokeVirtual(mRef));

            } catch (ByteCodeException | IOException | ResolveException ex) {
                Logger.getLogger(VariableInitializerVisitor.class.getName()).log(Level.SEVERE, null, ex);
            }

        }

    }
    /**
     * method write all stack items to instruction list 
     * @param env current environment
     * @param stack current infix stack
     * @param node 
     * @throws CompilerException 
     */
    public static void PutOperatorInstructions(ExpressionEnvironment env, Deque<InfixOpNode.InfixOps> stack, SyntaxNode node) throws CompilerException {
        while (!stack.isEmpty()) {
            InfixOpNode.InfixOps op = stack.pop();
            PutOperatorInstruction(env, op, node);
        }
    }
    /**
     * write one infix operator to instruction list
     * @param env current environment
     * @param op operator
     * @param node
     * @throws CompilerException 
     */
    public static void PutOperatorInstruction(ExpressionEnvironment env, InfixOpNode.InfixOps op, SyntaxNode node) throws CompilerException {
        try {
            Type t2 = env.popType();
            Type t1 = env.popType();
            
            TypeUtils utils = env.getStage3Env().getTypeUtils();
            FieldDescriptor f2 = utils.makeFieldDescriptor(t2);
            FieldDescriptor f1 = utils.makeFieldDescriptor(t1);
            if (f1.isVoid() || f2.isVoid()) {
                throw new CompilerException("cannot apply operators to void", node.getSymbol());
            }
            String s2 = f2.getClassInfo().getClassName();
            String s1 = f1.getClassInfo().getClassName();
            switch (op) {

                case BIT_OR:
//                    if (s1.equals("java/lang/Boolean") && s2.equals("java/lang/Boolean")) {
//                        MethodRefInfo booleanValue = MethodRefInfo.make("java/lang/Boolean", "booleanValue", FieldDescriptor.fromString("boolean"));
//                        MethodRefInfo valueOf = MethodRefInfo.make("java/lang/Boolean", "valueOf", "java/lang/Boolean", "boolean");
//                        env.insertCode(new InvokeVirtual(booleanValue));
//                        env.insertCode(new Swap());
//                        env.insertCode(new InvokeVirtual(booleanValue));
//                        env.insertCode(new IOr());
//                        env.insertCode(new InvokeStatic(valueOf));
//                        env.setResultType(new Type.ClassType(null, "java/lang/Boolean", null));
//                    } else {
//                        throw new CompilerException("incompatible types for operator |", node.getSymbol());
//                    }
//                    break;
                case BIT_AND:
//                    if (s1.equals("java/lang/Boolean") && s2.equals("java/lang/Boolean")) {
//                        MethodRefInfo booleanValue = MethodRefInfo.make("java/lang/Boolean", "booleanValue", FieldDescriptor.fromString("boolean"));
//                        MethodRefInfo valueOf = MethodRefInfo.make("java/lang/Boolean", "valueOf", "java/lang/Boolean", "boolean");
//                        env.insertCode(new InvokeVirtual(booleanValue));
//                        env.insertCode(new Swap());
//                        env.insertCode(new InvokeVirtual(booleanValue));
//                        env.insertCode(new IAnd());
//                        env.insertCode(new InvokeStatic(valueOf));
//                        env.setResultType(new Type.ClassType(null, "java/lang/Boolean", null));
//                    } else {
//                        throw new CompilerException("incompatible types for operator |", node.getSymbol());
//                    }
//                    break;
                case BIT_XOR:
//                    if (s1.equals("java/lang/Boolean") && s2.equals("java/lang/Boolean")) {
//                        MethodRefInfo booleanValue = MethodRefInfo.make("java/lang/Boolean", "booleanValue", FieldDescriptor.fromString("boolean"));
//                        MethodRefInfo valueOf = MethodRefInfo.make("java/lang/Boolean", "valueOf", "java/lang/Boolean", "boolean");
//                        env.insertCode(new InvokeVirtual(booleanValue));
//                        env.insertCode(new Swap());
//                        env.insertCode(new InvokeVirtual(booleanValue));
//                        env.insertCode(new IXOr());
//                        env.insertCode(new InvokeStatic(valueOf));
//                        env.setResultType(new Type.ClassType(null, "java/lang/Boolean", null));
//                    } else {
//                        throw new CompilerException("incompatible types for operator ^", node.getSymbol());
//                    }
//                    break;
                    throw new CompilerException("operator " + op + " not supported", node.getSymbol());
                case EQL: {
                    if (f1.isPrimitive() || f2.isPrimitive()) {
                        throw new CompilerException("primitive types not supported by ==");
                    }
                    //System.err.println(env.getOperandStack()); //<- Immer leer
                    MethodRefInfo info = MethodRefInfo.make("teajay/util/TJObjects", "equals", "java/lang/Boolean", "java/lang/Object", "java/lang/Object");
                    env.insertCode(new InvokeStatic(info));
                    env.setResultType(new Type.ClassType(null, "java/lang/Boolean", null));
                }
                break;
                case NEQL: {
                    if (f1.isPrimitive() || f2.isPrimitive()) {
                        throw new CompilerException("primitive types not supported by !=", node.getSymbol());
                    }
                    MethodRefInfo info = MethodRefInfo.make("teajay/util/TJObjects", "equals", "java/lang/Boolean", "java/lang/Object", "java/lang/Object");
                    env.insertCode(new InvokeStatic(info));
                    MethodRefInfo booleanValue = MethodRefInfo.make("java/lang/Boolean", "booleanValue", FieldDescriptor.fromString("boolean"));
                    env.insertCode(new InvokeVirtual(booleanValue));
                    env.insertCode(new IConst(1));
                    env.insertCode(new IXOr());
                    MethodRefInfo valueOf = MethodRefInfo.make("java/lang/Boolean", "valueOf", "java/lang/Boolean", "boolean");
                    env.insertCode(new InvokeStatic(valueOf));
                    env.setResultType(new Type.ClassType(null, "java/lang/Boolean", null));
                }
                break;
                case LESS:
                    putStaticOperator("operator_less", env, s1, s2, new Type.ClassType(null, "java/lang/Boolean", null), node);
                    break;
                case GREATER:
                    putStaticOperator("operator_greater", env, s1, s2, new Type.ClassType(null, "java/lang/Boolean", null), node);
                    break;
                case LESS_EQUAL:
                    putStaticOperator("operator_leq", env, s1, s2, new Type.ClassType(null, "java/lang/Boolean", null), node);
                    break;
                case GREATER_EQUAL:
                    putStaticOperator("operator_geq", env, s1, s2, new Type.ClassType(null, "java/lang/Boolean", null), node);
                    break;
                case ADD: {
                    if (s1.equals("java/lang/String") || s2.equals("java/lang/String")) {
                        MethodRefInfo concat = MethodRefInfo.make("java/lang/String", "concat", "java/lang/String", "java/lang/String");
                        MethodRefInfo valueOf1;
                        if (f1.isPrimitive()) {
                            valueOf1 = MethodRefInfo.make("java/lang/String", "valueOf", "java/lang/String", f1.getClassInfo().getClassName());
                        } else {
                            valueOf1 = MethodRefInfo.make("java/lang/String", "valueOf", "java/lang/String", "java/lang/Object");
                        }
                        MethodRefInfo valueOf2;
                        if (f2.isPrimitive()) {
                            valueOf2 = MethodRefInfo.make("java/lang/String", "valueOf", "java/lang/String", f2.getClassInfo().getClassName());
                        } else {
                            valueOf2 = MethodRefInfo.make("java/lang/String", "valueOf", "java/lang/String", "java/lang/Object");
                        }
                        env.insertCode(new InvokeStatic(valueOf2));
                        env.insertCode(new Swap());
                        env.insertCode(new InvokeStatic(valueOf1));
                        env.insertCode(new Swap());
                        env.insertCode(new InvokeVirtual(concat));
                        env.setResultType(new Type.ClassType(null, "java/lang/String", null));
                    } else {
                        putStaticOperator("operator_add", env, s1, s2, null, node);
                    }
                }
                break;
                case SUB:
                    putStaticOperator("operator_sub", env, s1, s2, null, node);
                    break;
                case MULT:
                    putStaticOperator("operator_mult", env, s1, s2, null, node);
                    break;
                case DIV:
                    putStaticOperator("operator_div", env, s1, s2, null, node);
                    break;
                case MOD:
                case LEFT_SHIFT:
                case RIGHT_SHIFT:
                case DOUBLE_RIGHT_SHIFT:
                    throw new CompilerException("unsupported operator: " + op, node.getSymbol());
                default:
                    throw new AssertionError(op.name());

            }
        } catch (ByteCodeException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
    }

    private static void putStaticOperator(String methodName, ExpressionEnvironment env, String s1, String s2, Type.ClassType returnType, SyntaxNode node) throws CompilerException {
        String caller = env.getStage3Env().getClassFile().getThis().getClassName();
        Pair<MethodInfo, MethodFinder.DistanceTailPair> pair = MethodFinder.findMethod(methodName, s1, caller, node, MethodFinder.QueryType.STATIC_ONLY,
                FieldDescriptor.fromString(s1), FieldDescriptor.fromString(s2));
        
        if (s1.equals("java/lang/String") && s2.equals("java/lang/String")) {
            pair = MethodFinder.findMethod(methodName, "teajay/runtime/StringType", caller, node, MethodFinder.QueryType.STATIC_ONLY,
                    FieldDescriptor.fromString(s1), FieldDescriptor.fromString(s2));
        } else if (pair == null) {
            pair = MethodFinder.findMethod(methodName, s2, caller, node, MethodFinder.QueryType.STATIC_ONLY,
                    FieldDescriptor.fromString(s1), FieldDescriptor.fromString(s2));
        } else {
            Pair<MethodInfo, MethodFinder.DistanceTailPair> tmp = MethodFinder.findMethod(methodName, s2, caller, node, MethodFinder.QueryType.STATIC_ONLY,
                    FieldDescriptor.fromString(s1), FieldDescriptor.fromString(s2));
            if (tmp != null) {
                int s2dist = tmp.value2.distance;
                int s1dist = pair.value2.distance;
                if (s2dist < s1dist) {
                    pair = tmp;
                }
            }
        }

        if (pair == null || (returnType != null && !pair.value1.getDescriptor().getReturnType().getClassInfo().getClassName().equals(returnType.name))) {
            throw new CompilerException("no such method: " + (returnType != null ? returnType : "?") + " " + methodName + "(" + s1 + "," + s2 + ")", node.getSymbol());
        } else if (pair.value1.isVarArgs()) {
            throw new CompilerException("operator methods may not be varargs", node.getSymbol());
        } else if (pair.value1.getDescriptor().getReturnType().isVoid()) {
            throw new CompilerException("operator methods must have a return value", node.getSymbol());
        }

        env.insertCode(new InvokeStatic(pair.value1.makeMethodRef()));
        Type retType = returnType;
        if (retType == null) {
            retType = TypeUtils.getMethodSignature(pair.value1).returnType;
            if (retType instanceof Type.SimpleType) {
                Type.SimpleType simpleType = (Type.SimpleType) retType;
                if (!simpleType.isPrimitiveType()) {
                    throw new CompilerException("generic static operators not supported yet", node.getSymbol());
                }
            }
            
        }
        env.setResultType(retType);
    }
    /**
     * 
     * @param env current expression environment
     * @param stack current infix operator stack
     * @param opPriority his priority stops inserting to instruction list
     * @param node
     * @throws CompilerException 
     */
    public static void PutOperatorInstructionsByPriority(ExpressionEnvironment env, Deque<InfixOpNode.InfixOps> stack, int opPriority, InfixOpNode node) throws CompilerException {
        while (!stack.isEmpty() && stack.peek().getPriority() >= opPriority) {
            InfixOpNode.InfixOps op = stack.pop();
            PutOperatorInstruction(env, op, node);
        }

    }
}
