/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package codegeneration.instruction;

import abstractsyntaxtree.ExpressionNode;
import abstractsyntaxtree.SyntaxNode;
import codegeneration.environment.ExpressionEnvironment;
import codegeneration.environment.Stage3Environment;
import datastructs.Pair;
import exceptions.CompilerException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import stage3.utils.InformationCenter;
import symbol.DerivationTree;
import symbol.FieldTypeResolver;
import symbol.LocalEntry;
import symbol.MethodResolver;
import teajay.bytecode.BaseType;
import teajay.bytecode.ClassFile;
import teajay.bytecode.attributes.code.AAStore;
import teajay.bytecode.attributes.code.ALoad;
import teajay.bytecode.attributes.code.ANewArray;
import teajay.bytecode.attributes.code.AStore;
import teajay.bytecode.attributes.code.CheckCast;
import teajay.bytecode.attributes.code.GetField;
import teajay.bytecode.attributes.code.GetStatic;
import teajay.bytecode.attributes.code.InvokeInterface;
import teajay.bytecode.attributes.code.InvokeStatic;
import teajay.bytecode.attributes.code.InvokeVirtual;
import teajay.bytecode.attributes.code.JVMInstruction;
import teajay.bytecode.attributes.code.JVMInstructionList;
import teajay.bytecode.attributes.code.Pop;
import teajay.bytecode.attributes.code.StackMapFrame;
import teajay.bytecode.attributes.code.Swap;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.cpool.InterfaceMethodRefInfo;
import teajay.bytecode.cpool.MethodRefInfo;
import teajay.bytecode.cpool.NameAndTypeInfo;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.fields.FieldInfo;
import teajay.bytecode.methods.MethodInfo;
import teajay.generics.check.TypeException;
import teajay.generics.check.TypeUtils;
import teajay.generics.parser.Type;
import static teajay.generics.parser.Type.WildcardType.Kind.UNBOUNDED;
import teajay.methods.MethodFinder;
import teajay.runtime.exceptions.ResolveException;
import teajay.util.QIDTools;
import teajay.util.io.ClassFileRef;
import teajay.util.io.ClassFinder;
import teajay.util.tjattribute.TeaJayAttribute;

/**
 *
 * @author Michael (Burzan | Kußmann)
 */
public class MethodInvokeCreator {

    public static class CallContext {

        private final Stage3Environment stage3Env;
        private final JVMInstruction instruction;
        private final StackMapFrame frame;
        private Type ownerClass;
        private String methodName;
        private final int popCnt;
        private boolean nonPublic = false;
        private boolean ignoreSynthetic = true;
        private boolean isStatic = false;

        public Stage3Environment getStage3Env() {
            return stage3Env;
        }

        public CallContext(Stage3Environment stage3Env, JVMInstruction instruction, StackMapFrame frame, Type ownerClass, String methodName, int popCnt) {
            this(stage3Env, instruction, frame, ownerClass, methodName, popCnt, true);
        }

        public CallContext(Stage3Environment stage3Env, JVMInstruction instruction, StackMapFrame frame, Type ownerClass, String methodName, int popCnt, boolean ignoreSynthetic) {
            this.stage3Env = stage3Env;
            this.instruction = instruction;
            this.frame = frame;
            this.ownerClass = ownerClass;
            this.methodName = methodName;
            this.popCnt = popCnt;
            this.ignoreSynthetic = ignoreSynthetic;

        }

        public CallContext(Stage3Environment stage3Env, Type ownerClass, String methodName, int popCnt) {
            this(stage3Env, ownerClass, methodName, popCnt, true);
        }

        public CallContext(Stage3Environment stage3Env, Type ownerClass, String methodName, int popCnt, boolean ignoreSynthetic) {
            this(stage3Env, ownerClass, methodName, popCnt, ignoreSynthetic, false);
        }

        public CallContext(Stage3Environment stage3Env, Type ownerClass, String methodName, int popCnt, boolean ignoreSynthetic, boolean isStatic) {
            this.instruction = null;
            this.frame = null;
            this.stage3Env = stage3Env;
            this.ownerClass = ownerClass;
            this.methodName = methodName;
            this.popCnt = popCnt;
            this.ignoreSynthetic = ignoreSynthetic;
            this.isStatic = isStatic;
        }

        public boolean isStatic() {
            return isStatic;
        }

        public int getPopCnt() {
            return this.popCnt;
        }

        public JVMInstruction getOverrideInstruction() {
            return this.instruction;
        }

        public StackMapFrame getStackMapFrame() {
            return this.frame;
        }

        public boolean hasAloadThis() {
            if (this.instruction == null) {
                return false;
            }
            return true;
        }

        public Type getOwnerClass() {
            return this.ownerClass;
        }

        public void overrideOwnerclass(Type ownerClass) {
            this.ownerClass = ownerClass;
        }

        public String getMethodName() {
            return this.methodName;
        }

        public void setNonPublic() {
            this.nonPublic = true;
        }

        public boolean isNonPublic() {
            return this.nonPublic;
        }

        public boolean doIgnoreSynthetic() {
            return ignoreSynthetic;
        }

        public void setMethodName(String name) {
            this.methodName = name;
        }
    }

    public static CallContext createMethodAccess(Stage3Environment stage3Env, ExpressionEnvironment env, String ident, List<String> selectorList, SyntaxNode node) throws ByteCodeException, CompilerException, IOException, ResolveException {
        return createMethodAccess(stage3Env, env, ident, selectorList, true, node);
    }

    public static CallContext createMethodAccess(Stage3Environment stage3Env, ExpressionEnvironment env, String ident, List<String> selectorList, boolean ignoreSynthetic, SyntaxNode node)
            throws ByteCodeException, CompilerException, IOException, ResolveException {
        return createMethodAccess(stage3Env, env, ident, selectorList, ignoreSynthetic, null, node);
    }

    public static CallContext createMethodAccess(Stage3Environment stage3Env, ExpressionEnvironment env, String ident, List<String> selectorList, boolean ignoreSynthetic, Type closureOverride, SyntaxNode node)
            throws ByteCodeException, CompilerException, IOException, ResolveException {
        if (selectorList == null || selectorList.isEmpty()) {
            //check is ident a closure in field
            LocalEntry entry = stage3Env.getLocalTable().getEntry(ident);
            if (entry != null) { // CLOSURE LOCAL TEST
                Type type = closureOverride != null ? closureOverride : entry.getType();

                if (type instanceof Type.ClassType) {
                    Type.ClassType classType = (Type.ClassType) type;
                    if (classType.name.equals("teajay/lang/Closure")) {
                        env.insertCode(new ALoad(entry.getIndex()));
                        env.setResultType(type);
                        return new CallContext(stage3Env, type, "execute", 1, ignoreSynthetic);
                    }

                }
            } //else {
            // Check is field a closure
            FieldInfo field = FieldTypeResolver.find(env.getFile(), ident);
            if (field != null && field.getDescriptor().getClassInfo().getClassName().equals("teajay/lang/Closure")) {
                if (!field.isStatic()) { // Non static closure call
                    env.insertCode(new ALoad(0));//Put this on stack
                    env.insertCode(new GetField(field.getFieldRefInfo(stage3Env.getClassFile().getThis()))); // Put field  on stack
                } else {
                    //static closure call
                    env.insertCode(new GetStatic(field.getFieldRefInfo(stage3Env.getClassFile().getThis()))); // Put static field on stack
                }
                Type fieldType = stage3Env.getTypeUtils().getFieldType(field);
                env.setResultType(fieldType);
                //Make closure call context 
                return new CallContext(stage3Env, fieldType, "execute", 1, ignoreSynthetic);
            } else {
                // THEN is my Ident a member call  or static call
                JVMInstruction thisCall = new ALoad(0);
                env.insertCode(thisCall);
                StackMapFrame frame = env.getMethod().getCode().getStackMapTable().getFrame(thisCall);
                env.setResultType(stage3Env.getThis());
                return new CallContext(stage3Env, thisCall, frame, stage3Env.getThis(), ident, 1, ignoreSynthetic);
            }
            //}
        }
        int size = selectorList.size();
        VariableInstructionCreator varInstructionCreator = new RValueVariableInstructionCreator(stage3Env);
        LocalEntry entry = stage3Env.getLocalTable().getEntry(ident);//SymbolFacade.instance().getLocalTable().getEntry(ident);
        if (entry != null) { //Is ident a local variable
            int popCnt = varInstructionCreator.createLocalVariableInstruction(entry, ident, selectorList.subList(0, size - 1), node);
            return new CallContext(stage3Env, env.typeOfResult(), selectorList.get(size - 1), popCnt, ignoreSynthetic);
        } else {
            FieldInfo field = FieldTypeResolver.find(env.getFile(), ident);
            if (field != null) {
                int popCnt = varInstructionCreator.createFieldInstruction(ident, selectorList.subList(0, size - 1), false, node);
                return new CallContext(stage3Env, env.typeOfResult(), selectorList.get(size - 1), popCnt, ignoreSynthetic);
            } else {
                //Check is ident a class name
                if (size == 1) {
                    String className = stage3Env.getSymbolFacade().getFullQualifiedId(ident);
                    if (className == null) {
                        throw new CompilerException("Class " + ident + " not found", node.getSymbol());
                    }
                    Type cType = new Type.ClassType(null, className, null);
                    return new CallContext(stage3Env, cType, selectorList.get(0), 0, ignoreSynthetic, true);
                }
                int popCnt;
                Type classType = isStaticClassCall(ident, selectorList.subList(0, size - 1), stage3Env);
                if (classType != null) {
                    popCnt = 0;
                    return new CallContext(stage3Env, classType, selectorList.get(size - 1), popCnt, ignoreSynthetic, true);
                } else {
                    popCnt = varInstructionCreator.createClassInstruction(ident, selectorList.subList(0, size - 1), node);
                }
                return new CallContext(stage3Env, env.typeOfResult(), selectorList.get(size - 1), popCnt, ignoreSynthetic);
            }
        }

    }

    public static void createMethodCall(CallContext context, ExpressionEnvironment env,
            String ident, List<String> selectorList, SyntaxNode node)
            throws ByteCodeException, CompilerException, IOException, ResolveException {
        createMethodCall(context, env, ident, selectorList, Collections.EMPTY_LIST, node);
    }

    public static void createMethodCall(CallContext context, ExpressionEnvironment env,
            String ident, List<String> selectorList, List<Type> typeArgs, SyntaxNode node)
            throws ByteCodeException, CompilerException, IOException, ResolveException {
        boolean isNonStaticCall = false;
        if (selectorList == null) {
            selectorList = Collections.emptyList();
        }
        if (ident != null && !selectorList.isEmpty()) {
            if (selectorList.get(0).equals("...")) { //Call from unknown selector
                isNonStaticCall = true;
            } else {
                isNonStaticCall = context.getStage3Env().getLocalTable().getEntry(ident) != null;
                isNonStaticCall |= FieldTypeResolver.isFieldInClassFile(context.getStage3Env().getClassFile(), ident);
            }
        }
        Type.ClassType ownerCType;
        if (context.getOwnerClass() instanceof Type.ClassType) {
            ownerCType = (Type.ClassType) context.getOwnerClass();
        } else if (context.getOwnerClass() instanceof Type.WildcardType) {
            Type.WildcardType wildcardType = (Type.WildcardType) context.getOwnerClass();
            switch (wildcardType.kind) {
                case EXTENDS:
                    if (wildcardType.boundType instanceof Type.ClassType) {
                        ownerCType = (Type.ClassType) wildcardType.boundType;
                    } else if (wildcardType.boundType instanceof Type.SimpleType) {
                        Type.SimpleType simpleType = (Type.SimpleType) wildcardType.boundType;
                        if (simpleType.isPrimitiveType()) {
                            throw new CompilerException("Cannot call methods on primitives");
                        }
                        ownerCType = context.getStage3Env().getTypeUtils().getUnerasedType(((Type.SimpleType) context.getOwnerClass()).name);
                    } else {
                        throw new CompilerException("Method calls on " + wildcardType.boundType + " not supported", node.getSymbol());
                    }
                    break;
                case SUPER:
                case UNBOUNDED:
                    ownerCType = new Type.ClassType(null, "java/lang/Object", null);
                    break;
                default:
                    throw new AssertionError(wildcardType.kind.name());

            }
        } else if (context.getOwnerClass() instanceof Type.SimpleType) {
            Type.SimpleType simpleType = (Type.SimpleType) context.getOwnerClass();
            if (simpleType.isPrimitiveType()) {
                throw new CompilerException("Cannot call methods on " + simpleType, node.getSymbol());
            }
            ownerCType = context.getStage3Env().getTypeUtils().getUnerasedType(simpleType.name);
        } else {
            throw new CompilerException("Method calls on " + context.getOwnerClass().getClass().getSimpleName() + " not supported", node.getSymbol());
        }
        //System.err.println(ident + selectorList + " - " + ownerCType);
        MethodInfo info;
        //XXX: Mapping String methods to StringType
        FieldDescriptor[] callParameters = env.getCallParameters().clone();
        List<Type> callTypeParameters = new ArrayList<>(callParameters.length);
        callTypeParameters.addAll(env.getCallTypeParameters());
        boolean overrideStatic = false;
        if (context.getMethodName().equals("hashCode") && callParameters.length == 0) {
            ownerCType = new Type.ClassType(null, "teajay/util/TJObjects", null);
            callTypeParameters.add(0, new Type.ClassType(null, "java/lang/Object", null));
            FieldDescriptor[] temp = new FieldDescriptor[callParameters.length + 1];
            temp[0] = FieldDescriptor.fromClass(Object.class);
            System.arraycopy(callParameters, 0, temp, 1, callParameters.length);
            callParameters = temp;
            overrideStatic = true;
        } else if (ownerCType.name.equals("java/lang/String") || ownerCType.name.equals("java/lang/CharSequence")
                || ownerCType.name.equals("java/io/File")) {
            ClassFileRef type;
            switch (ownerCType.name) {
                case "java/lang/String":
                    type = ClassFinder.instance().find("teajay/runtime/StringType");
                    break;
                case "java/lang/CharSequence":
                    type = ClassFinder.instance().find("teajay/runtime/CharSequenceType");
                    break;
                default:
                    type = ClassFinder.instance().find("teajay/runtime/FileType");
                    break;
            }
            if (type == null) {
                throw new Error("Corrupted teajay runtime environment");
            }
            ClassFile wrapType = type.get();

            if (wrapType.getMethods(context.getMethodName()) != null) {
                Type.ClassType oldType = new Type.ClassType(null, ownerCType.name, null);
                callTypeParameters.add(0, oldType);
                ownerCType = new Type.ClassType(null, type.getQID(), null);
                FieldDescriptor[] temp = new FieldDescriptor[callParameters.length + 1];
                temp[0] = FieldDescriptor.fromString(oldType.name);
                System.arraycopy(callParameters, 0, temp, 1, callParameters.length);
                callParameters = temp;
                overrideStatic = true;
            }
        } else if (ownerCType.name.equals("java/lang/System")) {
            ClassFileRef vmType = ClassFinder.instance().find("teajay/runtime/VM");
            if (vmType == null) {
                throw new Error("Corrupted teajay runtime environment");
            }
            ClassFile stringTypeClass = vmType.get();

            if (stringTypeClass.getMethods(context.getMethodName()) != null) {
                ownerCType = new Type.ClassType(null, "teajay/runtime/VM", null);
            }
        }

        TypeUtils.checkTypes(ownerCType, env.getStage3Env().getThis(), node.getSymbol());

        ClassFileRef find = ClassFinder.instance().find(ownerCType.name);
        ClassInfo ownerInfo = find.get().getThis();
        boolean isClosure = false;
        Type.MethodType closureSig = null;
        int tailLengthForVarArgs;
        if (ownerInfo != null && ownerInfo.getClassName().equals("teajay/lang/Closure") && selectorList.isEmpty()) {
            info = find.get().getMethods("execute").get(0);
            tailLengthForVarArgs = ownerCType.typeArgs.size() - 1;
            closureSig = new Type.MethodType(ownerCType.typeArgs.subList(1, ownerCType.typeArgs.size()), ownerCType.typeArgs.get(0));
            isClosure = true;
        } else {
            if (ownerInfo != null && ownerInfo.getClassName().equals("teajay/lang/Closure")) {
                ownerCType = new Type.ClassType(ownerCType.outerType, ownerCType.name, null);
            }
            try {
                Pair<MethodInfo, MethodFinder.DistanceTailPair> result = MethodFinder.findMethod(context.getMethodName(), Arrays.asList(callParameters), ownerCType.name, env.getStage3Env().getClassFile().getThis().getClassName(), node, MethodFinder.QueryType.ALL, context.doIgnoreSynthetic());
                if (result == null) {
                    List<ClassFileRef> staticImports = env.getStage3Env().getSymbolFacade().getStaticImports();
                    if (staticImports != null) {
                        //TODO: Static imports
                    }
                    throw new CompilerException("No such method in " + ownerCType.name + ": " + context.getMethodName() + " " + Arrays.toString(env.getCallParameters()));
                }
                if (result.value1.isStatic() && !overrideStatic && isNonStaticCall) {
                    throw new CompilerException("Cannot call static methods on instances", node.getSymbol());
                }
                if (!result.value1.isStatic() || overrideStatic) {
                    if (context.isStatic()) {
                        throw new CompilerException("Cannot call non-static method via type name");
                    }
                }

                info = result.value1;
                tailLengthForVarArgs = result.value2.tailLength;
            } catch (CompilerException ex) {
                throw new CompilerException(ex.getErrorMessage(), node.getSymbol());
            }
        }

        if (!isClosure && info.isPublic() && context.isNonPublic() && selectorList.isEmpty()) {
            throw new CompilerException("cannot call public method " + info.getName().getValue() + " without \"this.\"", node.getSymbol());
        }

        if (info.isVarArgs() || isClosure) {
            //Create array for vararg call
            JVMInstructionList ins = context.getStage3Env().getMethodInfo().getCode().getInstructionList();
            ins.pushInt(tailLengthForVarArgs);
            if (isClosure) {
                ins.add(new ANewArray(new ClassInfo("java/lang/Object")));
            } else {
                List<FieldDescriptor> params = info.getDescriptor().getParameters();
                FieldDescriptor varArgArray = params.get(params.size() - 1);
                ins.add(new ANewArray(varArgArray.getElementType().getClassInfo()));
            }
            int arrayIndex = context.getStage3Env().getLocalTable().getNextId();
            ins.add(new AStore(arrayIndex));
            for (int i = tailLengthForVarArgs - 1; i >= 0; i--) {
                ins.add(new ALoad(arrayIndex));
                ins.add(new Swap());
                ins.pushInt(i);
                ins.add(new Swap());
                ins.add(new AAStore());
            }
            ins.add(new ALoad(arrayIndex));
        }



        //Keep the stack consistent
        for (int i = 0; i < context.getPopCnt(); i++) {
            env.popType();
        }

        //  isMethodAccessPossible(info, context, env, find, node);

        MethodRefInfo callInfo;
        callInfo = MethodRefInfo.make(find.get().getThis(), context.getMethodName(),
                info.getDescriptor().getReturnType(),
                info.getDescriptor().getParameters().toArray(new FieldDescriptor[info.getDescriptor().getParameterCount()]));
        if (info.isStatic() && context.hasAloadThis() && !overrideStatic) {
            env.insertCode(new InvokeStatic(callInfo));
            //Pop aload_0
            if (!info.getDescriptor().getReturnType().isVoid()) {
                env.insertCode(new Swap());
            }
            env.insertCode(new Pop());
        } else {
            if (info.isStatic()) {
                env.insertCode(new InvokeStatic(callInfo));
            } else {
                if (find.get().isInterface()) {
                    env.insertCode(new InvokeInterface(new InterfaceMethodRefInfo(find.get().getThis(),
                            new NameAndTypeInfo(context.getMethodName(), info.getDescriptor()))));
                } else {
                    env.insertCode(new InvokeVirtual(callInfo));
                }
            }
        }

        Type.MethodType methodSignatureByEnv;
        if (isClosure) {
            methodSignatureByEnv = closureSig;
        } else {
            methodSignatureByEnv = context.getStage3Env().getTypeUtils().getMethodSignatureByEnv(info, ownerCType, typeArgs);
        }


        try {
            if (info.isVarArgs() && !isClosure) {
                List<Type> paramTypes = new ArrayList<>();
                paramTypes.addAll(methodSignatureByEnv.paramTypes.subList(0, methodSignatureByEnv.paramTypes.size() - 1));

                Type.ArrayType last = (Type.ArrayType) methodSignatureByEnv.paramTypes.get(methodSignatureByEnv.paramTypes.size() - 1);
                for (int i = 0; i < tailLengthForVarArgs; i++) {
                    paramTypes.add(last.elemType);
                }
                methodSignatureByEnv = new Type.MethodType(methodSignatureByEnv.typeParamTypes, paramTypes, methodSignatureByEnv.returnType, methodSignatureByEnv.throwsTypes);
            }
            context.getStage3Env().getTypeUtils().checkCallability(methodSignatureByEnv, callTypeParameters);
        } catch (TypeException ex) {
            //ex.printStackTrace(System.err);
            throw new CompilerException("Parameters missmatch (" + info + "): " + ex.getMessage(), node.getSymbol());
        }

        FieldDescriptor makeFieldDescriptor = context.stage3Env.getTypeUtils().makeFieldDescriptor(methodSignatureByEnv.returnType);
        env.resetCallParameters();

        if (methodSignatureByEnv.returnType instanceof Type.SimpleType) {
            Type.SimpleType simpleType = (Type.SimpleType) methodSignatureByEnv.returnType;
            if (!simpleType.isPrimitiveType()) {

                if (!makeFieldDescriptor.getClassInfo().getClassName().equals("java/lang/Object")) {
                    context.stage3Env.getMethodInfo().getCode().getInstructionList().add(new CheckCast(makeFieldDescriptor.getClassInfo()));
                    env.setResultType(methodSignatureByEnv.returnType);
                } else {
                    env.setResultType(methodSignatureByEnv.returnType);
                }
            } else {
                if (simpleType.name.equals("boolean") && !info.getDefiningClass().getClassName().equals("teajay/lang/java/JavaPrimitives")) {
                    //Do boolean wrap
                    context.stage3Env.getMethodInfo().getCode().getInstructionList().add(
                            new InvokeStatic(MethodRefInfo.make("java/lang/Boolean",
                            "valueOf", FieldDescriptor.fromClass(Boolean.class),
                            FieldDescriptor.fromBaseType(BaseType.Boolean))));
                    env.setResultType(new Type.ClassType(null, "java/lang/Boolean", null));
                } else {
                    env.setResultType(methodSignatureByEnv.returnType);
                }
            }
        } else {
            FieldDescriptor cType = info.getDescriptor().getReturnType();
            if (!cType.equals(makeFieldDescriptor)) {
                context.stage3Env.getMethodInfo().getCode().getInstructionList().add(new CheckCast(makeFieldDescriptor.getClassInfo()));
            }
            env.setResultType(methodSignatureByEnv.returnType);
        }
    }

    public static void makeConstructorCall(ExpressionEnvironment env, String location, boolean superCall, SyntaxNode node) throws CompilerException, ByteCodeException, IOException, ResolveException {
        ClassFileRef locationRef = ClassFinder.instance().find(location);
        if (locationRef == null) {
            throw new CompilerException("Class not found: " + location);
        }
        ClassFile locationFile = locationRef.get();

        Pair<MethodInfo, MethodFinder.DistanceTailPair> constructor = MethodFinder.findMethod("<init>", Arrays.asList(env.getCallParameters()), location, env.getClassName(), node, MethodFinder.QueryType.NON_VIRTUAL_NON_STATIC);
        if (constructor == null) {
            throw new CompilerException("No such constructor <init> " + env.getCallTypeParameters(), node.getSymbol());
        }
        MethodInfo info = constructor.value1;
        int tailLengthForVarArgs = constructor.value2.tailLength;
        JVMInstructionList ins = env.getMethod().getCode().getInstructionList();

        //TODO: Backport this to createMethodCall
        boolean lastIsArray = false;
        FieldDescriptor[] callParams = env.getCallParameters();
        if (callParams.length > 0 && info.getDescriptor().getParameterCount() > 0) {
            List<FieldDescriptor> params = info.getDescriptor().getParameters();
            FieldDescriptor varArgArray = params.get(params.size() - 1);

            FieldDescriptor last = callParams[callParams.length - 1];
            if (last.isArray()) {
                ClassFileRef refa = ClassFinder.instance().find(varArgArray.getClassInfo().getClassName());
                ClassFileRef refb = ClassFinder.instance().find(last.getClassInfo().getClassName());
                lastIsArray = DerivationTree.instance().isAssignable(refa, refb);
            }
        }

        Type.MethodType methodSignatureByEnv = env.getStage3Env().getTypeUtils().getMethodSignatureByEnv(info, env.getStage3Env().getThis(), new Type.ClassType(null, location, null), null);

        if (info.isVarArgs() && !(lastIsArray && tailLengthForVarArgs == 0)) {
            ins.pushInt(tailLengthForVarArgs);
            List<FieldDescriptor> params = info.getDescriptor().getParameters();
            FieldDescriptor varArgArray = params.get(params.size() - 1);
            ins.add(new ANewArray(varArgArray.getElementType().getClassInfo()));
            int arrayIndex = env.getStage3Env().getLocalTable().getNextId();
            ins.add(new AStore(arrayIndex));
            for (int i = tailLengthForVarArgs - 1; i >= 0; i--) {
                ins.add(new ALoad(arrayIndex));
                ins.add(new Swap());
                ins.pushInt(i);
                ins.add(new Swap());
                ins.add(new AAStore());
            }
            ins.add(new ALoad(arrayIndex));

            List<Type> paramTypes = new ArrayList<>();
            paramTypes.addAll(methodSignatureByEnv.paramTypes.subList(0, methodSignatureByEnv.paramTypes.size() - 1));

            Type.ArrayType last = (Type.ArrayType) methodSignatureByEnv.paramTypes.get(methodSignatureByEnv.paramTypes.size() - 1);
            for (int i = 0; i < tailLengthForVarArgs; i++) {
                paramTypes.add(last.elemType);
            }
            methodSignatureByEnv = new Type.MethodType(methodSignatureByEnv.typeParamTypes, paramTypes, methodSignatureByEnv.returnType, methodSignatureByEnv.throwsTypes);
        }

        env.getStage3Env().getTypeUtils().checkCallability(methodSignatureByEnv, env.getCallTypeParameters());
        TeaJayAttribute.Type type = InformationCenter.getTeaJayAttributeType(locationFile);
        switch (type) {
            case Impl:
                if (superCall) {
                    throw new CompilerException("Super calls on implmentations are not allowed.", node.getSymbol());
                }
            //env.insertCode(new InvokeSpecial(info.makeMethodRef()));
            //break;
            case Class:
            case TJBuiltinType:
            case Hidden:
            //env.insertCode(new InvokeSpecial(info.makeMethodRef()));
            //break;
            case Typedef:
                env.getStage3Env().invokeDynamic(env.getMethod(), new ClassInfo(location), info.getDescriptor().getParameters().toArray(new FieldDescriptor[0]));
                break;
            case Enum:
            case Closure:
            case Interface:
                throw new CompilerException("Cannot instatiate enums, closures or interfaces.", node.getSymbol());
            default:
                throw new AssertionError(type.name());

        }
    }

    @Deprecated
    private static void makeIterator(Stage3Environment stage3Env, ExpressionNode node) throws CompilerException {
        try {
            String methodName = "iterator";
            Type callerType = stage3Env.getExpressionEnvironment().typeOfResult();
            MethodInfo info = MethodResolver.existMethod(stage3Env, callerType,
                    methodName);
            MethodRefInfo makeMethodRef = info.makeMethodRef();
            ClassFile find = ClassFinder.instance().find(((Type.ClassType) callerType).name).get();
            if (find.isInterface()) {
                InterfaceMethodRefInfo interfaceRefInfo = new InterfaceMethodRefInfo(find.getThis(),
                        new NameAndTypeInfo(methodName, info.getDescriptor()));

                stage3Env.getExpressionEnvironment().insertCode(new InvokeInterface(interfaceRefInfo));
            } else {
                stage3Env.getExpressionEnvironment().insertCode(new InvokeVirtual(makeMethodRef));
            }
        } catch (ByteCodeException | IOException | ResolveException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
    }

    @Deprecated
    private static void isMethodAccessPossible(MethodInfo info, CallContext context, ExpressionEnvironment env, ClassFileRef ownerRef, SyntaxNode node) throws CompilerException, ByteCodeException, IOException, ResolveException {
        if (info.isPrivate()) {
            ClassFile owner = ownerRef.get();
            //Allow closures to access private methods of their defining class?
            if (!owner.getThis().equals(env.getFile().getThis())/* && !isInnerClass(owner, env.getFile())*/) {
                throw new CompilerException("Access to private method " + context.getMethodName() + " is not allowed", node.getSymbol());
            }
        } else if (info.isPackagePrivate()) {
            if (!QIDTools.samePackage(ownerRef.get().getThis().getClassName(), env.getClassName())) {
                throw new CompilerException("Access to package private method " + context.getMethodName() + " is not allowed", node.getSymbol());
            }
        } else if (info.isProtected()) {
            if (!DerivationTree.instance().isAssignable(ownerRef, ClassFinder.instance().find(env.getClassName()))) {
                throw new CompilerException("Access to protected method " + context.getMethodName() + " is not allowed", node.getSymbol());
            }
        }
    }

    private static Type isStaticClassCall(String ident, List<String> identifierList, Stage3Environment stage3Env) {
        for (int i = 0; i < identifierList.size(); i++) {
            ident += "/" + identifierList.get(i);
        }
        if (ClassFinder.instance().find(ident) != null) {
            return new Type.ClassType(null, ident, null);
        }
        return null;
    }
}
