/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package codegeneration.instruction;

import abstractsyntaxtree.IdentifierSuffixNode;
import abstractsyntaxtree.SyntaxNode;
import codegeneration.environment.ExpressionEnvironment;
import codegeneration.environment.Stage3Environment;
import exceptions.CompilerException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import symbol.DerivationTree;
import symbol.FieldTypeResolver;
import symbol.LocalEntry;
import teajay.bytecode.ClassFile;
import teajay.bytecode.attributes.code.ALoad;
import teajay.bytecode.attributes.code.GetField;
import teajay.bytecode.attributes.code.GetStatic;
import teajay.bytecode.attributes.code.PutField;
import teajay.bytecode.attributes.code.PutStatic;
import teajay.bytecode.attributes.code.Swap;
import teajay.bytecode.cpool.FieldRefInfo;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.fields.FieldInfo;
import teajay.runtime.exceptions.ResolveException;
import teajay.util.QIDTools;
import teajay.util.io.ClassFileRef;
import teajay.util.io.ClassFinder;

/**
 *
 * @author Michael Burzan Base class for Variable Access
 */
public abstract class VariableInstructionCreator {

    protected final Stage3Environment stage3Env;
    protected final ExpressionEnvironment env;

    public VariableInstructionCreator(Stage3Environment stage3Env) {
        this.stage3Env = stage3Env;
        this.env = stage3Env.getExpressionEnvironment();
    }

    /**
     * create variable access: Type.name
     *
     * @param ident
     * @param identifierList
     * @param node the current traversing syntax node
     * @return number of types that can be removed from stack
     * @throws CompilerException
     * @throws ByteCodeException
     * @throws IOException
     * @throws ResolveException
     */
    public abstract int createClassInstruction(String ident, List<String> identifierList, SyntaxNode node) throws CompilerException, ByteCodeException, IOException, ResolveException;

    /**
     * create field access: a.b
     *
     * @param fieldName name of the field a
     * @param selectors name of the selectors a
     * @param isArrayAccess compatibility to java
     * @param node the current traversing syntax node
     * @return number of types that can be removed from stack
     * @throws CompilerException
     * @throws ByteCodeException
     * @throws IOException
     * @throws ResolveException
     */
    public abstract int createFieldInstruction(String fieldName, List<String> selectors, boolean isArrayAccess, SyntaxNode node) throws CompilerException, ByteCodeException, IOException, ResolveException;

    /**
     * Access to local variable: a.b
     *
     * @param entry entry of the local variable in the local table
     * @param ident name of the variable a
     * @param selector b
     * @param node
     * @return number of types that can be removed from stack
     * @throws CompilerException
     * @throws ByteCodeException
     * @throws IOException
     * @throws ResolveException
     */
    public abstract int createLocalVariableInstruction(LocalEntry entry, String ident, List<String> selector, SyntaxNode node) throws CompilerException, ByteCodeException, IOException, ResolveException;

    /**
     * create access to local variable by suffix type a[]
     *
     * @param entry
     * @param ident
     * @param selector
     * @param node
     * @param suffixType
     * @return number of types that can be removed from stack
     * @throws ByteCodeException
     * @throws CompilerException
     * @throws IOException
     * @throws ResolveException
     */
    public abstract int createLocalVariableInstruction(LocalEntry entry, String ident, List<String> selector, IdentifierSuffixNode node, IdentifierSuffixNode.SuffixType suffixType) throws ByteCodeException, CompilerException, IOException, ResolveException;

    /**
     *
     * @param ident
     * @param identifierList
     * @return package.Type.identifier => Type.identifier
     */
    protected List<String> flatten(String ident, List<String> identifierList) {
        String fullQualifiedId = stage3Env.getSymbolFacade().getFullQualifiedId(ident);
        if (fullQualifiedId != null) {
            ArrayList<String> list = new ArrayList<>();
            list.add(fullQualifiedId);
            list.addAll(identifierList);
            return list;
        } else {

            //String s = ident;
            for (int i = 0; i < identifierList.size(); i++) {
                ident += "/" + identifierList.get(i);
                //is ident a class file
                if (ClassFinder.instance().find(ident) != null) {
                    ArrayList<String> list = new ArrayList<>();
                    list.add(ident);
                    list.addAll(identifierList.subList(i + 1, identifierList.size()));

                    return list;
                }
            }
        }

        return null;
    }

    protected ClassFile testField(ClassFile ownerClass, String fieldName, String fromPackage, ExpressionEnvironment env, SyntaxNode node) throws CompilerException, ByteCodeException, IOException, ResolveException {
        FieldInfo field = FieldTypeResolver.find(ownerClass, fieldName);
        //field does not exist
        if (field == null) {
            throw new CompilerException("Field " + fieldName + "in class " + ownerClass.getThis().getSimpleName() + "does not exist", node.getSymbol());
        }
        ClassFile fieldClassFile = ClassFinder.instance().find(field.getDescriptor().getClassInfo().getClassName()).get();
        // It is not allowed to access from a variable to a static field (TeaJay property)
        if (field.isStatic()) {
            throw new CompilerException("Acces to static field " + fieldName + " from variable is not allowed", node.getSymbol());
        }
        if (field.isPrivate()) {
            if (!ownerClass.getThis().equals(env.getFile().getThis())) {
                throw new CompilerException("Access to private field " + fieldName + "  is not allowed", node.getSymbol());
            }
        }
        if (field.isPackagePrivate()) {
            String fieldPackage = QIDTools.packageName(ownerClass.getThis().getClassName());
            if (!QIDTools.samePackage(fromPackage, fieldPackage)) {
                throw new CompilerException("Access to package private field " + fieldName + "  is not allowed from other package", node.getSymbol());
            }
        }
        if (field.isProtected()) {
            ClassFileRef ownerRef = ClassFinder.instance().find(ownerClass.getThis().getClassName());
            ClassFileRef ref = ClassFinder.instance().find(env.getClassName());
            if (!DerivationTree.instance().isDerivated(ownerRef, ref)) {
                throw new CompilerException("Access to protected field " + fieldName + " is not allowed because " + ownerClass.getThis().getSimpleName() + "is not base of " + ownerClass.getThis().getSimpleName(), node.getSymbol());
            }
        }
        return fieldClassFile;
    }

    protected int accessToField(Stage3Environment stage3Env, String className, List<String> selectors, SyntaxNode node) throws CompilerException, ByteCodeException, IOException, ResolveException {
        return accessToField(stage3Env, className, selectors, false, node);
    }

    protected int accessToField(Stage3Environment stage3Env, String className, List<String> selectors, boolean noAload0, SyntaxNode node) throws CompilerException, ByteCodeException, IOException, ResolveException {
        ExpressionEnvironment env = stage3Env.getExpressionEnvironment();
        ClassFileRef accessClass = ClassFinder.instance().find(className);
        ClassFile accessingClass = accessClass.get();
        ClassFile iter = accessingClass;
        int pushCnt = 0;
        for (int i = 0; i < selectors.size() - 1; i++) {
            ClassFile next = testField(iter, selectors.get(i), QIDTools.packageName(className), env, node);
            FieldInfo field = FieldTypeResolver.find(iter, selectors.get(i));
            FieldRefInfo ref = FieldRefInfo.make(iter.getThis(), selectors.get(i), field.getDescriptor());
            env.insertCode(new GetField(ref));
            env.setResultType(stage3Env.getTypeUtils().getFieldType(field));
            iter = next;
            pushCnt++;
        }
        testField(iter, selectors.get(selectors.size() - 1), QIDTools.packageName(className), env, node);
        FieldInfo field = FieldTypeResolver.find(iter, selectors.get(selectors.size() - 1));
        FieldRefInfo ref = FieldRefInfo.make(iter.getThis(), selectors.get(selectors.size() - 1), field.getDescriptor());
        if (env.isLvalue()) {
            if (!field.isStatic()) {
                if (!noAload0) {
                    env.addLValueInstruction(new ALoad(0), stage3Env.getThis());
                    env.setLValueInstruction(new Swap());
                }
            }
            if (field.isFinal()) {
                throw new CompilerException(field.getFieldName() + " is final", node.getSymbol());
            }
            env.addLValueInstruction(field.isStatic() ? new PutStatic(ref) : new PutField(ref), stage3Env.getTypeUtils().getFieldType(field));
        } else {
            if (!field.isStatic()) {
                if (!noAload0) {
                    env.insertCode(new ALoad(0));
                }
                env.insertCode(new GetField(ref));
            } else {
                env.insertCode(new GetStatic(ref));
            }
            pushCnt++;

            env.setResultType(stage3Env.getTypeUtils().getFieldType(field));

        }
        return pushCnt;
    }
}
