/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package codegeneration.javavisitor.classes;

import codegeneration.abstractvisitors.AbstractNormalClassDeclarationVisitor;
import codegeneration.abstractvisitors.AbstractClassDeclarationVisitor;
import abstractsyntaxtree.EnumDeclarationNode;
import abstractsyntaxtree.ModifierNode;
import abstractsyntaxtree.ModifierNode.Modifier;
import static abstractsyntaxtree.ModifierNode.Modifier.ABSTRACT;
import static abstractsyntaxtree.ModifierNode.Modifier.FINAL;
import static abstractsyntaxtree.ModifierNode.Modifier.PUBLIC;
import static abstractsyntaxtree.ModifierNode.Modifier.STRICTFP;
import abstractsyntaxtree.NormalClassDeclarationNode;
import codegeneration.Visitor;
import codegeneration.javavisitor.enums.EnumDeclarationVisitor;
import java.util.ArrayList;
import java.util.List;
import exceptions.CompilerException;
import java.io.IOException;
import symbol.SymbolFacade;
import teajay.bytecode.BaseType;
import teajay.bytecode.ClassAccessFlag;
import teajay.bytecode.ClassFile;
import teajay.bytecode.attributes.code.*;
import teajay.bytecode.attributes.signature.Signature;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.cpool.FieldRefInfo;
import teajay.bytecode.cpool.MethodDescriptor;
import teajay.bytecode.cpool.MethodRefInfo;
import teajay.bytecode.cpool.NameAndTypeInfo;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.fields.FieldAccessFlag;
import teajay.bytecode.fields.FieldInfo;
import teajay.bytecode.methods.MethodAccessFlag;
import teajay.bytecode.methods.MethodInfo;
import teajay.runtime.exceptions.DuplicateQualifiedIdentifierException;
import teajay.runtime.exceptions.ResolveException;
import teajay.util.io.ClassFinder;

/**
 *
 * @author cefour
 */
public class ClassDeclarationVisitor extends AbstractClassDeclarationVisitor<ClassFile> {

    private List<ModifierNode.Modifier> modList;
    private ClassFile classFile;
    private AbstractNormalClassDeclarationVisitor normalClassDeclarationVisitor;
    private final String packageName;
    private EnumDeclarationVisitor enumDeclVisitor;

    public ClassDeclarationVisitor(List<Modifier> modList, String packageName) {
        this.modList = modList;
        this.packageName = packageName;
    }

    @Override
    /**
     * Method creates the ClassFile object. The full qualified class name
     * consists of the package name plus the name of the class. The method
     * checks the modifier on correctness for a class declaration and uniqueness
     */
    public void begin(NormalClassDeclarationNode node) {
        genClassFile(node.getClassName(), false);
    }

    @Override
    public Visitor visit(NormalClassDeclarationNode node) {
        return new NormalClassDeclarationVisitor(classFile);
    }

    @Override
    /**
     * If the source code has not a constructor then generate this method the
     * standard constructor. Created code attribute for standard is: aload_0
     * invokespecial
     * <init>:()V return
     *
     */
    public void end(NormalClassDeclarationNode node) {
        genDefaultConstructor(false);
    }

    @Override
    public void begin(EnumDeclarationNode node) {
        genClassFile(node.getName(), true);
        try {
            classFile.addField(new FieldInfo("$VALUES", FieldDescriptor.makeArray(FieldDescriptor.fromClassFile(classFile), 1),
                    FieldAccessFlag.PRIVATE, FieldAccessFlag.STATIC, FieldAccessFlag.FINAL, FieldAccessFlag.SYNTHETIC));
        } catch (ByteCodeException ex) {
            ex.printStackTrace(System.err);
        }
    }

    @Override
    public Visitor visit(EnumDeclarationNode node) {
        this.enumDeclVisitor = new EnumDeclarationVisitor(classFile);
        return enumDeclVisitor;
    }

    @Override
    public void end(EnumDeclarationNode node) {
        try {
            List<String> enumConstants = enumDeclVisitor.getContainer();
            for (String name : enumConstants) {
                classFile.addField(new FieldInfo(name, FieldDescriptor.fromClassFile(classFile),
                        FieldAccessFlag.PUBLIC, FieldAccessFlag.STATIC, FieldAccessFlag.FINAL, FieldAccessFlag.ENUM));
            }
            //values()
            {
                MethodInfo values = new MethodInfo("values",
                        new MethodDescriptor(FieldDescriptor.makeArray(FieldDescriptor.fromClassFile(classFile), 1)),
                        classFile.getThis(), MethodAccessFlag.PUBLIC, MethodAccessFlag.STATIC);
                classFile.addMethod(values);
                JVMInstructionList lst = values.getCode().getInstructionList();
                FieldRefInfo ref = new FieldRefInfo(classFile.getThis(),
                        new NameAndTypeInfo<>("$VALUES", FieldDescriptor.makeArray(FieldDescriptor.fromClassFile(classFile), 1)));
                lst.add(new GetStatic(ref));
                ClassInfo array = FieldDescriptor.makeArray(FieldDescriptor.fromClassFile(classFile), 1).getClassInfo();
                MethodRefInfo mref = new MethodRefInfo(array, new NameAndTypeInfo<>("clone",
                        new MethodDescriptor(FieldDescriptor.fromClass(Object.class))));
                lst.add(new InvokeVirtual(mref));
                lst.add(new CheckCast(array));
                lst.add(new Return(ReturnType.Reference));
            }
            //end values()

            //valueOf(String)
            {
                MethodInfo values = new MethodInfo("valueOf",
                        new MethodDescriptor(FieldDescriptor.fromClassFile(classFile), FieldDescriptor.fromClass(String.class)),
                        classFile.getThis(), MethodAccessFlag.PUBLIC, MethodAccessFlag.STATIC);
                classFile.addMethod(values);
                JVMInstructionList lst = values.getCode().getInstructionList();
                lst.add(new Ldc(classFile.getThis()));
                lst.add(new ALoad(0));
                MethodRefInfo mref = new MethodRefInfo(ClassInfo.fromClass(Enum.class),
                        new NameAndTypeInfo<>("valueOf", new MethodDescriptor(FieldDescriptor.fromClass(Enum.class),
                        FieldDescriptor.fromClass(Class.class), FieldDescriptor.fromClass(String.class))));
                lst.add(new InvokeStatic(mref));
                lst.add(new CheckCast(classFile.getThis()));
                lst.add(new Return(ReturnType.Reference));
            }
            //end valueOf(String)
            {
                MethodInfo clinit = new MethodInfo("<clinit>",
                        new MethodDescriptor(FieldDescriptor.getVoid()),
                        classFile.getThis(), MethodAccessFlag.STATIC);
                classFile.addMethod(clinit);
                JVMInstructionList lst = clinit.getCode().getInstructionList();

                //MethodDescriptor baseConstr = new MethodDescriptor(FieldDescriptor.getVoid(),
                //        FieldDescriptor.fromClass(String.class), FieldDescriptor.fromBaseType(BaseType.Int));
                //NameAndTypeInfo nameInfo = new NameAndTypeInfo("<init>", baseConstr);
                //MethodRefInfo init = new MethodRefInfo(this.classFile.getThis(), nameInfo);
                MethodRefInfo init = MethodRefInfo.make(this.classFile.getThis(), "<init>", FieldDescriptor.getVoid(),
                        FieldDescriptor.fromClass(String.class), FieldDescriptor.fromBaseType(BaseType.Int));

                int i = 0;
                for (String name : enumConstants) {
                    lst.add(new New(classFile.getThis()));
                    lst.add(new Dup());
                    lst.add(new Ldc(name));
                    lst.pushInt(i);
                    lst.add(new InvokeSpecial(init));
                    FieldRefInfo finfo = new FieldRefInfo(classFile.getThis(),
                            new NameAndTypeInfo<>(name, FieldDescriptor.fromClassFile(classFile)));
                    lst.add(new PutStatic(finfo));
                    i++;
                }
                lst.pushInt(i);
                lst.add(new ANewArray(classFile.getThis()));
                lst.add(new Dup());
                i = 0;
                for (String name : enumConstants) {
                    lst.pushInt(i);
                    FieldRefInfo finfo = new FieldRefInfo(classFile.getThis(),
                            new NameAndTypeInfo<>(name, FieldDescriptor.fromClassFile(classFile)));
                    lst.add(new GetStatic(finfo));
                    lst.add(new AAStore());
                    i++;
                    if (i < enumConstants.size()) {
                        lst.add(new Dup());
                    }
                }
                FieldRefInfo ref = new FieldRefInfo(classFile.getThis(),
                        new NameAndTypeInfo<>("$VALUES", FieldDescriptor.makeArray(FieldDescriptor.fromClassFile(classFile), 1)));
                lst.add(new PutStatic(ref));
                lst.add(new Return());
            }
            genDefaultConstructor(true);
        } catch (ByteCodeException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
    }

    @Override
    public ClassFile getClassContainer() {
        return this.classFile;
    }

    private void genDefaultConstructor(boolean isEnum) {
        if (!isEnum) {
            if (classFile.getMethods("<init>") == null) {
                try {
                    MethodDescriptor baseConstructor = new MethodDescriptor(FieldDescriptor.getVoid());
                    NameAndTypeInfo objectInit = new NameAndTypeInfo("<init>", baseConstructor);

                    MethodRefInfo ref = new MethodRefInfo(this.classFile.getSuper(), objectInit);

                    MethodDescriptor desc = new MethodDescriptor(FieldDescriptor.getVoid());
                    MethodInfo method = new MethodInfo("<init>", desc, classFile.getThis(), MethodAccessFlag.PUBLIC);

                    this.classFile.addMethod(method);
                    method.getCode().getInstructionList().add(new ALoad(0));
                    method.getCode().getInstructionList().add(new InvokeSpecial(ref));
                    method.getCode().getInstructionList().add(new Return());

                } catch (ByteCodeException e) {
                    e.printStackTrace(System.err);
                    System.exit(-1);
                }
            }
        } else if (classFile.getMethods("<init>") == null) {
            try {
                MethodDescriptor baseConstr = new MethodDescriptor(FieldDescriptor.getVoid(),
                        FieldDescriptor.fromClass(String.class), FieldDescriptor.fromBaseType(BaseType.Int));
                MethodInfo minfo = new MethodInfo("<init>", baseConstr, classFile.getThis(), MethodAccessFlag.PRIVATE);
                classFile.addMethod(minfo);
                minfo.addAttribute(new Signature("()V"));
                NameAndTypeInfo init = new NameAndTypeInfo("<init>", baseConstr);
                MethodRefInfo mref = new MethodRefInfo(this.classFile.getSuper(), init);
                JVMInstructionList lst = minfo.getCode().getInstructionList();
                lst.add(new ALoad(0));
                lst.add(new ALoad(1));
                lst.add(new ILoad(2));
                lst.add(new InvokeSpecial(mref));
                lst.add(new Return());
            } catch (ByteCodeException e) {
                e.printStackTrace(System.err);
                System.exit(-1);
            }
        }
    }

    private void genClassFile(String simpleName, boolean isEnum) {
        String className;
        if (this.packageName != null) {
            className = this.packageName + "/" + simpleName;
        } else {
            //Anonymous package
            className = simpleName;
        }
        try {

            ArrayList<ClassAccessFlag> flagList = new ArrayList<>();
            boolean[] isRepeatedModifier = {false, false, false, false};
            for (ModifierNode.Modifier mod : modList) {
                switch (mod) {
                    case ABSTRACT:
                        if (!isRepeatedModifier[0]) {
                            flagList.add(ClassAccessFlag.ABSTRACT);
                            isRepeatedModifier[0] = true;
                        } else {
                            throw new CompilerException("Repeated modifier "
                                    + mod.name() + " in class declaration " + className);
                        }
                        break;
                    case FINAL:
                        if (isEnum) {
                            throw new CompilerException("modifier final not allowed for enums in " + className);
                        }
                        if (!isRepeatedModifier[1]) {
                            flagList.add(ClassAccessFlag.FINAL);
                            isRepeatedModifier[1] = true;

                        } else {
                            throw new CompilerException("Repeated modifier "
                                    + mod.name() + " in class declaration " + className);
                        }
                        break;
                    case PUBLIC:
                        if (!isRepeatedModifier[2]) {
                            flagList.add(ClassAccessFlag.PUBLIC);
                            isRepeatedModifier[2] = true;
                        } else {
                            throw new CompilerException("Repeated modifier "
                                    + mod.name() + " in class declaration " + className);
                        }
                        break;
                    case STRICTFP:
                        throw new CompilerException("modifier strictfp not allowed for class, in " + className);
                    /*if (!isRepeatedModifier[3]) {
                     isRepeatedModifier[3] = true;
                     } else {
                     throw new CompilerException("Repeated modifier "
                     + mod.name() + " in class declaration " + className);
                     }
                     break;*/
                    default:
                        throw new CompilerException("Modifier" + mod.name()
                                + " not allowed in class declaration " + className);

                }

            }
            flagList.add(ClassAccessFlag.SUPER);
            this.classFile = new ClassFile(className, flagList.toArray(new ClassAccessFlag[0]));
            if (isEnum) {
                flagList.add(ClassAccessFlag.ENUM);
                flagList.add(ClassAccessFlag.FINAL);
                classFile.setSuper("java/lang/Enum");
                classFile.addAttribute(new Signature("Ljava/lang/Enum<L" + classFile.getThis().getClassName() + ";>;"));
            }

            /*SET CLASS FILE AND DISTRIBUTE TO SYSTEM*/
            ClassFinder.instance().addClassFile(classFile);
//            SymbolFacade.instance().addNameFromImport(classFile.getThis().getClassName());


        } catch (ByteCodeException | CompilerException | DuplicateQualifiedIdentifierException  e) {
            e.printStackTrace(System.err);
            System.exit(-1);

        }
    }
}
