package lexer;
import exceptions.LexerException;
import static lexer.UTF16.utf16Decode;
/**
 * A lexer for the Java programming language (1.7) and TeaJay.
 * @author Michael Kußmann <michi@outx.de>
 */
%%
%public
%final
%class Lexer
%yylexthrow LexerException
%unicode
%type Symbol
%column
%line
%javanl

%eofval{
int state = yystate();
if (state == TRADITIONAL_COMMENT) {
    error("unclosed comment", errorLine, errorColumn);
} else if (state == STRING || state == STRING_UNICODE_ESCAPE) {
    error("unclosed string literal", errorLine, errorColumn);
} else if (state == CHAR || state == CHAR_UNICODE_ESCAPE || state == CHAR_END) {
    error("unclosed character literal", errorLine, errorColumn);
} else {
    return symbol(Sym.EOF);
}
%eofval}

%init{
in = new UnicodeFilteredReader(in, ZZ_BUFFERSIZE + 16);
%init}

%{
    private int jumpback = YYINITIAL;
    private StringBuilder string = new StringBuilder(32);
    private Character utf16BlockTmp = null;
    private int errorLine = -1;
    private int errorColumn = -1;
    private String sourceFile = null;

    public void beginTemplate() {
        yybegin(TEMPLATE);
    }

    public void endTemplate() {
        yybegin(YYINITIAL);
    }

    private Symbol symbol(int type) {
        return new Symbol(type, yyline, yycolumn);
    }

    private Symbol symbol(int type, String value) {
        return new Symbol(type, yyline, yycolumn, value);
    }

    public void setSourceFile(String filename) {
        this.sourceFile = filename;
        if (zzReader instanceof UnicodeFilteredReader) {
            ((UnicodeFilteredReader) zzReader).setSourceFile(filename);
        }
    }

    private void error(String msg) throws LexerException {
        if (sourceFile != null) {
            throw new LexerException(msg, sourceFile, yyline + 1, yycolumn + 1);
        } else {
            throw new LexerException(msg, yyline + 1, yycolumn + 1);
        }
    }

    private void error(String msg, int line, int column) throws LexerException {
        if (sourceFile != null) {
            throw new LexerException(msg, sourceFile, line + 1, column + 1);
        } else {
            throw new LexerException(msg, line + 1, column + 1);
        }
    }

    private static String formatHex(char c) {
        String hex = Integer.toHexString(c).toUpperCase();
        if (c <= 0x000F) {
            return "U+000" + hex;
        } else if (c <= 0x00FF) {
            return "U+00" + hex;
        } else if (c <= 0x0FFF) {
            return "U+0" + hex;
        } else {
            return "U+" + hex;
        }
    }

    /**
     * This is only needed for 100% java compliance.
     */
    private void checkJavaIdentifier(String id) {
        char[] chars = id.toCharArray();
        if (chars.length < 1) {
            error("empty identifier");
        }
        int start = 1;
        if (!Character.isJavaIdentifierStart(chars[0])) {
            if (chars.length > 1) {
                Integer cp = utf16Decode(chars[0], chars[1]);
                if (cp != null && !Character.isJavaIdentifierStart(cp)) {
                    error("invalid character: " + chars[0] + " (" + formatHex(chars[0]) + ")");
                }
                start++;
            } else {
                error("invalid character: " + chars[0] + " (" + formatHex(chars[0]) + ")");
            }
        }
        for (int i = start; i < chars.length; i++) {
            if (!Character.isJavaIdentifierPart(chars[i])) {
                if (chars.length > i + 1) {
                    Integer cp = utf16Decode(chars[i], chars[i + 1]);
                    if (cp != null && !Character.isJavaIdentifierPart(cp)) {
                        error("invalid character: " + chars[i] + " (" + formatHex(chars[i]) + ")");
                    }
                    i++;
                } else {
                    error("invalid character: " + chars[i] + " (" + formatHex(chars[i]) + ")");
                }
            }
        }
    }
%}

LineTerminator = \r | \n | \r\n
Whitespace = {LineTerminator} | [ \t\f]

OneLineComment = "//" [^\r\n]*

FourByteChar = [\uD800-\uDBFF][\uDC00-\uDFFF]
SimpleIdentifier = [:jletter:][:jletterdigit:]*
Identifier = ([:jletter:] | {FourByteChar})([:jletterdigit:] | {FourByteChar})*

HexDigit = [0-9a-fA-F]
HexIntegerLiteral = 0[xX]({HexDigit}+
                    | {HexDigit}({HexDigit}|_)+{HexDigit})
OctIntegerLiteral = 0([0-7]+ | [0-7][0-7_]+[0-7])
DecIntegerLiteral = 0 | [1-9][0-9]* | [1-9][0-9_]+[0-9]
BinIntegerLiteral = 0b([01]+ | [01][01_]+[01])

IntegerLiteral = {DecIntegerLiteral}
                 | {HexIntegerLiteral}
                 | {OctIntegerLiteral}
                 | {BinIntegerLiteral}
FloatingPointLiteral = {DecFloatingPointLiteral} | {HexFloatingPointLiteral}

DecDigits = [0-9][0-9]* | [0-9][0-9_]+[0-9]
DecFloatingPointLiteral = {DecDigits}"."{DecDigits}?{ExponentPart}?{FloatTypeSuffix}?
                          | "."{DecDigits}{ExponentPart}?{FloatTypeSuffix}?
                          | {DecDigits}{ExponentPart}{FloatTypeSuffix}?
                          | {DecDigits}{ExponentPart}?{FloatTypeSuffix}
ExponentPart = [eE][+-]?{DecDigits}
FloatTypeSuffix = [fFdD]

HexDigits = {HexDigit}+
HexFloatingPointLiteral = {HexSignificand}{BinaryExponent}{FloatTypeSuffix}?
HexSignificand = {HexIntegerLiteral}
                 | {HexIntegerLiteral}"."
                 | 0[xX]{HexDigits}?"."{HexDigits}
BinaryExponent = [pP][+-]?{DecDigits}

OctalEscape = \\([0-3][0-7]{2} | [0-7]{1,2})

%state STRING
%state CHAR
%state TEMPLATE
%state CHAR_UNICODE_ESCAPE
%state STRING_UNICODE_ESCAPE
%state CHAR_END
%state TRADITIONAL_COMMENT

%%

<YYINITIAL, TEMPLATE> {
/* begin comments */
    "/*"            {
                        jumpback = yystate();
                        errorLine = yyline;
                        errorColumn = yycolumn;
                        yybegin(TRADITIONAL_COMMENT);
                    }
    {OneLineComment}
                    { }
/* begin keywords */
    "byte"          { return symbol(Sym.BYTE); }
    "short"         { return symbol(Sym.SHORT); }
    "char"          { return symbol(Sym.CHAR); }
    "int"           { return symbol(Sym.INT); }
    "long"          { return symbol(Sym.LONG); }
    "float"         { return symbol(Sym.FLOAT); }
    "double"        { return symbol(Sym.DOUBLE); }
    "boolean"       { return symbol(Sym.BOOLEAN); }
    "abstract"      { return symbol(Sym.ABSTRACT); }
    "void"          { return symbol(Sym.VOID); }
    "break"         { return symbol(Sym.BREAK); }
    "if"            { return symbol(Sym.IF); }
    "else"          { return symbol(Sym.ELSE);}
    "("             { return symbol(Sym.BRACKET_OPEN); }
    ")"             { return symbol(Sym.BRACKET_CLOSE); }
    "{"             { return symbol(Sym.CURLY_OPEN); }
    "}"             { return symbol(Sym.CURLY_CLOSE); }
    "["             { return symbol(Sym.EDGECLAMPS_OPEN); }
    "]"             { return symbol(Sym.EDGECLAMPS_CLOSE); }
    "implements"    { return symbol(Sym.IMPLEMENTS); }
    "while"         { return symbol(Sym.WHILE); }
    "do"            { return symbol(Sym.DO); }
    "this"          { return symbol(Sym.THIS); }
    "super"         { return symbol(Sym.SUPER); }
    "return"        { return symbol(Sym.RETURN); }
    "package"       { return symbol(Sym.PACKAGE);}
    "public"        { return symbol(Sym.PUBLIC); }
    "private"       { return symbol(Sym.PRIVATE); }
    "new"           { return symbol(Sym.NEW); }
    "import"        { return symbol(Sym.IMPORT); }
    "for"           { return symbol(Sym.FOR); }
    "extends"       { return symbol(Sym.EXTENDS); }
    ";"             { return symbol(Sym.SEMICOLON); }
    ":"             { return symbol(Sym.COLON); }
    ","             { return symbol(Sym.COMMA); }
    "try"           { return symbol(Sym.TRY); }
    "catch"         { return symbol(Sym.CATCH); }
    "continue"      { return symbol(Sym.CONTINUE); }
    "throw"         { return symbol(Sym.THROW); }
    "throws"        { return symbol(Sym.THROWS); }
/* begin TeaJay keywords */
    "Closure"       { return symbol(Sym.CLOSUREDEC); }
    "closure"       { return symbol(Sym.CLOSUREDEF);}
    "typeswitch"    { return symbol(Sym.TYPESWITCH); }
    "typedef"       { return symbol(Sym.TYPEDEF); }
    "typeimpl"      { return symbol(Sym.TYPEIMPL); }
    "interfaces"    { return symbol(Sym.INTERFACES);}
/* end TeaJay keywords */
    "static"        { return symbol(Sym.STATIC); }
/*    "switch"        { return symbol(Sym.SWITCH); } */
    "case"          { return symbol(Sym.CASE); }
    "default"       { return symbol(Sym.DEFAULT); }
    "enum"	    { return symbol(Sym.ENUM); }
    "."{Whitespace}*"*"
                    { return symbol(Sym.DOTSTAR); }
    "."             { return symbol(Sym.DOT); }
    "..."           { return symbol(Sym.TRIPLEDOT); }
    "interface"     { return symbol(Sym.INTERFACE); }
    "class"         { return symbol(Sym.CLASS); }
    "final"         { return symbol(Sym.FINAL); }
/*    "finally"       { return symbol(Sym.FINALLY);} */
    "native"        { return symbol(Sym.NATIVE); }
    "synchronized"  { return symbol(Sym.SYNCHRONIZED); }
    "transient"     { return symbol(Sym.TRANSIENT); }
    "strictfp"      { return symbol(Sym.STRICTFP);  }
    "volatile"      { return symbol(Sym.VOLATILE);}
/*    "protected"     { return symbol(Sym.PROTECTED); } */
    "assert"        { return symbol(Sym.ASSERT);    }
/*    "instanceof"    { return symbol(Sym.INSTANCEOF); } */
/*    "goto"          { return symbol(Sym.GOTO); }
    "const"         { return symbol(Sym.CONST); } */

/* begin literals */
    "true"|"false"  { return symbol(Sym.BOOL_L, yytext()); }
    "null"          { return symbol(Sym.NULL_L, "null"); }
    {FloatingPointLiteral}
                    {   char end = yycharat(yylength() - 1);
                        if (end =='f' || end == 'F') {
                            return symbol(Sym.FLOAT_L, yytext());
                        } else {
                            return symbol(Sym.DOUBLE_L, yytext());
                        }    
                    }

    {IntegerLiteral}[Ll]
                    { return symbol(Sym.LONG_L, yytext()); }
    {IntegerLiteral}
                    { return symbol(Sym.INT_L, yytext()); }

    \"              {   jumpback = yystate();
                        errorLine = yyline;
                        errorColumn = yycolumn;
                        string.setLength(0);
                        yybegin(STRING);
                    }
    \'              {   jumpback = yystate();
                        errorLine = yyline;
                        errorColumn = yycolumn;
                        string.setLength(0);
                        yybegin(CHAR);
                    }

/* begin operators */
    "!="            { return symbol(Sym.OP_NEQ); }
    "!"             { return symbol(Sym.OP_NOT); }
    "=="            { return symbol(Sym.OP_EQEQ); }
    "="             { return symbol(Sym.OP_EQ); }
    "++"            { return symbol(Sym.OP_PLUS_PLUS); }
    "+="            { return symbol(Sym.OP_PLUS_SET); }
    "+"             { return symbol(Sym.OP_PLUS); }
    "--"            { return symbol(Sym.OP_SUB_SUB); }
    "-"             { return symbol(Sym.OP_SUB); }
    "-="            { return symbol(Sym.OP_SUB_SET); }
    "*="            { return symbol(Sym.OP_MULT_SET); }
    "*"             { return symbol(Sym.OP_MULT); }
    "/="            { return symbol(Sym.OP_DIV_SET); }
    "/"             { return symbol(Sym.OP_DIV); }
    "&="            { return symbol(Sym.OP_AND_SET); }
    "&&"            { return symbol(Sym.AND); }
    "&"             { return symbol(Sym.OP_AND); }
    "|="            { return symbol(Sym.OP_OR_SET); }
    "||"            { return symbol(Sym.OR); }
    "|"             { return symbol(Sym.OP_OR); }
    "^="            { return symbol(Sym.OP_XOR_SET); }
    "^"             { return symbol(Sym.OP_XOR); }
    "%="            { return symbol(Sym.OP_MOD_SET); }
    "%"             { return symbol(Sym.OP_MOD); }
    "~"             { return symbol(Sym.OP_NEG); }
    "@"             { return symbol(Sym.AT); }
    "?"             { return symbol(Sym.QUESTIONMARK); }
}

<YYINITIAL> {
    ">>>="          { return symbol(Sym.OP_RIGHT_RIGHT_SHIFT_SET); }
    ">>>"           { return symbol(Sym.OP_RIGHT_RIGHT_SHIFT); }
    ">>="           { return symbol(Sym.OP_RIGHT_SHIFT_SET); }
    ">>"            { return symbol(Sym.OP_RIGHT_SHIFT); }
    ">="            { return symbol(Sym.OP_GEQ); }
    ">"             { return symbol(Sym.OP_GREATER); }
    "<<="           { return symbol(Sym.OP_LEFT_SHIFT_SET); }
    "<<"            { return symbol(Sym.OP_LEFT_SHIFT); }
    "<="            { return symbol(Sym.OP_LEQ); }
    "<"             { return symbol(Sym.OP_LESS); }
}

<TEMPLATE> {
    "<"             { return symbol(Sym.OP_LESS); }
    ">"             { return symbol(Sym.OP_GREATER); }
}

<YYINITIAL, TEMPLATE> {
    {Whitespace}    { }

    /* There shouldn't be any of these left */
    \\u+            { error("illegal unicode escape"); }
    {SimpleIdentifier}
                    { return symbol(Sym.IDENTIFIER, yytext()); }

    {Identifier}    { String txt = yytext(); checkJavaIdentifier(txt); return symbol(Sym.IDENTIFIER, txt); }
    [^]             { char c = yycharat(0); error("illegal character " + c + " (" + formatHex(c) + ")"); }
}

<STRING> {
    \"              { yybegin(jumpback); utf16BlockTmp = null; return symbol(Sym.STRING_L,string.toString()); }

    [^\n\r\"\\]+    { string.append(zzBuffer, zzStartRead, zzMarkedPos - zzStartRead); utf16BlockTmp = null; }

    \\b             { string.append('\b'); utf16BlockTmp = null; }
    \\t             { string.append('\t'); utf16BlockTmp = null; }
    \\r             { string.append('\r'); utf16BlockTmp = null; }
    \\n             { string.append('\n'); utf16BlockTmp = null; }
    \\f             { string.append('\f'); utf16BlockTmp = null; }
    \\u+            { yybegin(STRING_UNICODE_ESCAPE); }
    \\\'            { string.append('\''); utf16BlockTmp = null; }
    \\\"            { string.append('\"'); utf16BlockTmp = null; }
    \\\\            { string.append('\\'); utf16BlockTmp = null; }
    
    {OctalEscape}   { string.append((char) Integer.parseInt(yytext().substring(1), 8)); utf16BlockTmp = null; }

    \\              { utf16BlockTmp = null; error("illegal escape character", yyline, yycolumn + 1); }
    [\r\n]          { utf16BlockTmp = null; error("unclosed string literal"); }
}

<STRING_UNICODE_ESCAPE> {
    {HexDigit}{4}   { yybegin(STRING);
                        char code = (char) Integer.parseInt(yytext(), 16);
                        Integer utf32;
                        if (utf16BlockTmp != null && (utf32 = utf16Decode(utf16BlockTmp, code)) != null) {
                            string.deleteCharAt(string.length() - 1);
                            string.appendCodePoint(utf32);
                            utf16BlockTmp = null;
                        } else {
                            utf16BlockTmp = code;
                            string.append(code);
                        }
                    }

    [^]             { yybegin(STRING); error("illegal unicode escape"); }
}

<CHAR_UNICODE_ESCAPE> {
    {HexDigit}{4}   { yybegin(CHAR_END); string.append((char) Integer.parseInt(yytext(), 16)); }

    [^]             { yybegin(CHAR_END); error("illegal unicode escape"); }
}

<CHAR> {
    \'              { yybegin(jumpback); error("empty character literal"); }
    [^\n\r\'\\]     { yybegin(CHAR_END); string.append(yycharat(0)); }

    \\b             { yybegin(CHAR_END); string.append('\b'); }
    \\t             { yybegin(CHAR_END); string.append('\t'); }
    \\r             { yybegin(CHAR_END); string.append('\r'); }
    \\n             { yybegin(CHAR_END); string.append('\n'); }
    \\f             { yybegin(CHAR_END); string.append('\f'); }
    \\u+            { yybegin(CHAR_UNICODE_ESCAPE); }
    \\\'            { yybegin(CHAR_END); string.append('\''); }
    \\\"            { yybegin(CHAR_END); string.append('\"'); }
    \\\\            { yybegin(CHAR_END); string.append('\\'); }

    {OctalEscape}   { yybegin(CHAR_END);
                      string.append((char) Integer.parseInt(yytext().substring(1), 8));
                    }
    \\              { error("illegal escape character"); }
    [\r\n]          { error("illegal line end in character literal"); }
}

<CHAR_END> {
                    /* char is treated like a signed 16-bit integer literal */
    \'              { yybegin(jumpback); return symbol(Sym.CHAR_L, String.valueOf((int) string.charAt(0))); }
    [^]             { yybegin(jumpback); error("unclosed character literal"); }
}

<TRADITIONAL_COMMENT> {
    "*/"            { yybegin(jumpback); }
    [^]             { }
}