/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package minicompiler;

import teajay.util.tjattribute.TeaJayAttribute;
import exceptions.CompilerException;
import exceptions.LexerException;
import teajay.runtime.exceptions.ResolveException;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import teajay.bytecode.exceptions.ByteCodeException;
import java.io.PrintStream;
import teajay.bytecode.attributes.AttributesReader;
import teajay.bytecode.exceptions.MalformedAttributeException;
import teajay.generics.check.TypeException;
import teajay.runtime.exceptions.DuplicateQualifiedIdentifierException;
import teajay.util.io.ClassFinder;
import teajay.util.io.ProgressToggler;
import teajay.util.io.TJSourceFinder;
import teajay.util.tjattribute.TJBridge;

/**
 *
 * @author Michael Burzan, Michael Kußmann
 *
 *
 * Description: Ein Compilervorgang sieht wie folgt aus : 1. Eingabe eines
 * TJ-Files oder Dateiliste 2. Datei wird geparst und anhand von ihren
 * abhängigkeiten wird eine Queue gefüllt um die nächsten Dateien zu Compilieren
 * 2.1 ACHTUNG Unterverzeichnisse müssen mit beachtet werden. 3. Erzeugung der
 * .class Files geschieht im gleichen Verzeichniss wo die tj-Files sind ( Falls
 * Flag es nicht anders umleitet )
 *
 *
 */
public class Main {

    private static final String version = "0.1";

    private static void help() {
        System.err.println("Usage: tjc <files> <args> \n");
        System.err.println("   or: tjc <args> <files> \n");

        System.err.println("Arguments:");
        System.err.println("-cp <classpath>\t\tspecify the class path. Default is \".\" ");
        System.err.println("\tThe format is: <jar file or directory>{" + System.getProperty("path.separator", ":") + "<jar file or directory>}");
        System.err.println("-d <path>\t\tspecify the output directory. Default is \".\" ");
        System.err.println("-sourcepath <path>\tspecify the source path. Default is \".\" ");
        System.err.println("-encoding <encoding>\tspecify the input encoding. Default is " + System.getProperty("file.encoding"));
        System.err.println("-help");
        System.err.println("-version\t\tdisplay version information about the TeaJay compiler");
    }
    private static boolean debug;

    private static boolean readArgsAndSetup(String... args) throws CompilerException {
        // tjc test.tj -d [directory] -cp [path:path:...] -sourcepath [path] -encoding[code] -help -version  
        int length = args.length;
        if (length == 0) {
            help();
            return false;
        }
        CompileManager manager = CompileManager.instance();
        for (int i = 0; i < length; i++) {
            switch (args[i]) {
                case "-cp":
                case "-classpath":
                    if ((i + 1) < length) {
                        System.setProperty("tjc_cp", args[++i]);
                    } else {
                        help();
                        return false;
                    }
                    break;
                case "-debug":
                    debug = true;
                    System.setProperty("tjc_debug", "true");
                    break;
                case "-d":
                    if ((i + 1) < length) {
                        System.setProperty("tjc_d", args[++i]);
                    } else {
                        help();
                        return false;
                    }
                    break;
                case "-sourcepath":
                    if ((i + 1) < length) {
                        System.setProperty("tjc_sourcepath", args[++i]);
                    } else {
                        help();
                        return false;
                    }
                    break;
                case "-encoding":
                    if ((i + 1) < length) {
                        System.setProperty("tjc_encoding", args[++i]);
                    } else {
                        help();
                        return false;
                    }
                    break;
                case "-help":
                    help();
                    System.exit(0);
                    return false;
                case "-version":
                    version(System.err);
                    System.exit(0);
                    return false;
                default:
                    manager.addSourceFile(new File(args[i]));
            }
        }
        return true;
    }

    private static void version(PrintStream out) {
        out.println("TeaJay compiler version " + version);
        out.println("License: GPLv3 <http://www.gnu.org/licenses/gpl.html>");
        out.println("Creators of the TeaJay language: Michael Burzan and Michael Kußmann");
    }

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) throws FileNotFoundException, MalformedAttributeException, ByteCodeException, ResolveException {
        System.setOut(new PrintStream(
                System.getProperty("os.name").toLowerCase().contains("windows") ? "nul" : "/dev/null"));
        AttributesReader.registerAttribute(TeaJayAttribute.class);
        AttributesReader.registerAttribute(TJBridge.class);
        long start = System.nanoTime();
        try {
            if (readArgsAndSetup(args)) {
                Thread io = new Thread(new Runnable() {
                    @Override
                    public void run() {
                        ClassFinder.instance().addPaths(System.getProperty("sun.boot.class.path"), System.getProperty("java.class.path"),
                                CompileManager.instance().classPath());
                        TJSourceFinder.instance().addPaths(CompileManager.instance().sourcePath());
                    }
                });
                io.setUncaughtExceptionHandler(new Thread.UncaughtExceptionHandler() {
                    @Override
                    public void uncaughtException(Thread t, Throwable ex) {
                        if (ex instanceof DuplicateQualifiedIdentifierException) {
                            System.err.println("Error:");
                            System.err.print('\t');
                            System.err.println("duplicate qualified identifier encountered: " + ex.getMessage());
                        } else {
                            ex.printStackTrace(System.err);
                        }
                        System.exit(1);
                    }
                });
                io.start();
                io.join(5 * 1000);
                ProgressToggler toggler = new ProgressToggler();
                if (io.isAlive()) {
                    System.err.println("Try setting a less crowded class or source path for faster startup");
                }
                while (io.isAlive()) {
                    System.err.print("\rScanning classpath and source path " + toggler.toggle());
                    io.join(1500);
                }
                if (toggler.hasToggled()) {
                    System.err.println(" done.");
                }                
                Compiler compiler = new Compiler();
                compiler.compile();
                System.err.print("\tBuild completed in ");
                System.err.println((System.nanoTime() - start) / 1000000000 + "s.");

            } else {
                System.exit(1);
            }
        } catch (CompilerException | LexerException | TypeException | IOException ex) {
            if (!debug) {
                System.err.println("Error:");
                System.err.print('\t');
                System.err.println(ex.getMessage());
            } else {
                ex.printStackTrace(System.err);
            }
            System.exit(1);
        } catch (InterruptedException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
    }
}
