/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package stage3;

import abstractsyntaxtree.BlockNode;
import abstractsyntaxtree.ParExpressionNode;
import codegeneration.Visitor;
import codegeneration.abstractvisitors.AbstractDoStatementVisitor;
import codegeneration.environment.ExpressionEnvironment;
import codegeneration.environment.Stage3Environment;
import exceptions.CompilerException;
import stage3.states.BlockState;
import stage3.states.EndState;
import stage3.states.LoopState;
import teajay.bytecode.attributes.code.IMode;
import teajay.bytecode.attributes.code.If0;
import teajay.bytecode.attributes.code.InvokeVirtual;
import teajay.bytecode.attributes.code.Label;
import teajay.bytecode.attributes.code.NOP;
import teajay.bytecode.attributes.code.PlaceholderInstruction;
import teajay.bytecode.cpool.MethodRefInfo;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.generics.check.TypeException;
import teajay.generics.parser.Type;

/**
 *
 * @author Michael Burzan
 */
public class Stage3DoStatementVisitor extends AbstractDoStatementVisitor {

    private final Stage3Environment stage3Env;
    private Stage3BlockVisitor blockVisitor;
    private Label breakLabel;
    private Label continueLabel;
    private Label blockBegin;
    private boolean isEndState = false;
    private NOP nop = new NOP();
    public Stage3DoStatementVisitor(Stage3Environment stage3Env) {
        this.stage3Env = stage3Env;
    }

    @Override
    public void begin(BlockNode node) throws CompilerException {
        BlockState state = new BlockState();
        stage3Env.getLocalTable().addFrame();
        stage3Env.changeState(state, node);
        this.breakLabel = new Label("do while break label");
        this.continueLabel = new Label("do while continue label");

        LoopState loopState = stage3Env.hasState(LoopState.class);
        loopState.setBreakLabel(this.breakLabel);
        loopState.setLoopLabel(this.continueLabel);
        try{
        stage3Env.getLocalTable().addStackMapFrame(
            stage3Env.getExpressionEnvironment().getMethod().getCode().getInstructionList().add(nop));
        }catch(ByteCodeException ex){
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
    }

    @Override
    public void begin(ParExpressionNode node) throws CompilerException {
        LoopState lState = stage3Env.hasState(LoopState.class);
        if (!isEndState || lState.hasAccessToContinueLabel() || lState.hasAccessToBreakLabel()) {
            if (isEndState) {
                stage3Env.prevState(); // forget endstate
                stage3Env.prevState(); // forget blockstate
                stage3Env.getLocalTable().removeFrame();
            }
            if(lState.hasAccessToContinueLabel() || !isEndState){
                ExpressionEnvironment env = stage3Env.getExpressionEnvironment();
                BlockState state = stage3Env.hasState(BlockState.class);
                state.addLabel(continueLabel);
                env.insertCode(new NOP());
            }
        } else {
            throw new CompilerException("Unreachable statement in paranthesis expression", node.getSymbol());
        }
    }

    @Override
    public Visitor visit(BlockNode node) throws CompilerException {
        return blockVisitor = new Stage3BlockVisitor(stage3Env);
    }

    @Override
    public Visitor visit(ParExpressionNode node) throws CompilerException {
        LoopState lState = stage3Env.hasState(LoopState.class);
        if (this.isEndState && !lState.hasAccessToContinueLabel()) {
            return null;
        }
        return new Stage3ParExpressionVisitor(stage3Env);
    }

    @Override
    public void end(BlockNode node) throws CompilerException {
        if (blockVisitor.isEmptyBlock()) {
            stage3Env.getExpressionEnvironment().insertCode(new NOP());
        }
        EndState endState = stage3Env.hasState(EndState.class);
        if (endState != null) {
            this.isEndState = true;
        } else {
            stage3Env.getLocalTable().removeFrame();
            stage3Env.prevState();
        }
    }

    @Override
    public void end(ParExpressionNode node) throws CompilerException {
        LoopState lState = stage3Env.hasState(LoopState.class);
        /*
         * Instruction after the do block only possible, 
         * if the state machine not in end state or
         * in the block exist continue or break.
         */
        if (!isEndState || lState.hasAccessToContinueLabel() || lState.hasAccessToBreakLabel()) {
            ExpressionEnvironment env = stage3Env.getExpressionEnvironment();
            if (isEndState && !lState.hasAccessToContinueLabel()) {
                BlockState state = stage3Env.hasState(BlockState.class);
                state.addLabel(breakLabel);
                env.insertCode(new NOP());
            } else {
                Type.ClassType bType = new Type.ClassType(null, "java/lang/Boolean", null);
                Type typeOfResult = env.typeOfResult();
                try {
                    stage3Env.getTypeUtils().checkAssignability(bType, typeOfResult);
                    MethodRefInfo ref = MethodRefInfo.make("java/lang/Boolean", "booleanValue", "boolean");
                    env.getMethod().getCode().getInstructionList().add(new InvokeVirtual(ref));
                    env.getMethod().getCode().getInstructionList().add(new If0(IMode.ne, nop));
                    BlockState state = stage3Env.hasState(BlockState.class);
                    state.addLabel(breakLabel);
                    env.insertCode(new NOP());
                }catch (ByteCodeException ex){
                    ex.printStackTrace(System.err);
                    System.exit(-1);
                } 
                catch (TypeException ex) {
                    throw new CompilerException("do statetement expects Boolean", node.getSymbol());
                }
            }
        }


    }

    public Label getLabel() {
        LoopState loopState = stage3Env.hasState(LoopState.class);
        return loopState.getBreakLabel();
    }
}
