/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package symbol;

import teajay.runtime.exceptions.ResolveException;
import java.io.IOException;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import teajay.bytecode.ClassFile;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.fields.FieldInfo;
import teajay.util.io.ClassFileRef;
import teajay.util.io.ClassFinder;

/**
 *
 * @author Michael Burzan
 */
public class FieldTypeResolver {

    /**
     * 
     * @param file
     * @param varName
     * @return true exist field in class file otherwise false
     */
    public static boolean isFieldInClassFile(ClassFile file, String varName) {
        FieldInfo field = file.getField(varName);
        if (field != null) {
            return true;
        }
        return false;
    }
    /**
     * 
     * @param file class file where field is possible declared
     * @param varName name of the field
     * @return the class file where field is declared
     * @throws ByteCodeException
     * @throws IOException
     * @throws ResolveException 
     */
    public static ClassFile findOwner(ClassFile file, String varName) throws ByteCodeException, IOException, ResolveException {
        return findOwner(file, varName, new HashSet<ClassInfo>());
    }
    /**
     * 
     * @param file class file where field is declared
     * @param varName name of the field
     * @param visited visited class infos
     * @return the class file where the field is declared
     * @throws ByteCodeException
     * @throws IOException
     * @throws ResolveException 
     */
    public static ClassFile findOwner(ClassFile file, String varName, Set<ClassInfo> visited) throws ByteCodeException, IOException, ResolveException {
        FieldInfo field = file.getField(varName);
        if (field == null) {
            return findOwnerInSuper(file, varName, visited);
        }
        return file;
    }

    private static ClassFile findOwnerInSuper(ClassFile file, String varName, Set<ClassInfo> visited)
            throws ByteCodeException, IOException, ResolveException {
        ClassFile iter = file;
        //TODO WARNING: Warnung kann hier ausgegeben werden falls eine Variable 
        // verdeckt wird
        do {
            //Java compatibility: check in interfaces 
            Collection<ClassInfo> ifaces = file.getInterfaces();
            for (ClassInfo iface : ifaces) {
                if (!visited.contains(iface)) {
                    ClassFileRef ref = ClassFinder.instance().find(iface.getClassName());
                    FieldInfo candidate = find(ref.get(), varName, visited);
                    if (candidate != null) {
                        return ref.get();
                    }
                    visited.add(iface);
                }
            }
            if (iter.getSuper() == null) {
                return null;
            }
            //Java compatibility: check in parent classes
            ClassFileRef ref = ClassFinder.instance().find(iter.getSuper().getClassName());
            iter = ref.get();
            FieldInfo field = iter.getField(varName);
            if (field != null && !field.isPrivate()) {
                return ref.get();
            }
        } while (iter.getSuper() != null);


        return null;
    }

    public static FieldInfo find(ClassFile file, String varName) throws ByteCodeException, IOException, ResolveException {
        return find(file, varName, new HashSet<ClassInfo>());
    }

    public static FieldInfo find(ClassFile file, String varName, Set<ClassInfo> visited) throws ByteCodeException, IOException, ResolveException {
        FieldInfo field = file.getField(varName);
        if (field == null) {
            field = searchInSuper(file, varName, visited);
        }
        return field;
    }

    static private FieldInfo searchInSuper(ClassFile file, String varName, Set<ClassInfo> visited) throws ByteCodeException, IOException, ResolveException {
        ClassFile iter = file;
        if (iter.getSuper() == null) {
            return null;
        }
        //TODO WARNING: Warnung kann hier ausgegeben werden falls eine Variable 
        // verdeckt wird
        do {
            Collection<ClassInfo> ifaces = file.getInterfaces();
            for (ClassInfo iface : ifaces) {
                if (!visited.contains(iface)) {
                    ClassFileRef ref = ClassFinder.instance().find(iface.getClassName());
                    FieldInfo candidate = find(ref.get(), varName, visited);
                    if (candidate != null) {
                        return candidate;
                    }
                    visited.add(iface);
                }
            }
            ClassFileRef ref = ClassFinder.instance().find(iter.getSuper().getClassName());
            iter = ref.get();
            FieldInfo field = iter.getField(varName);
            if (field != null && !field.isPrivate()) {
                return field;
            }
        } while (iter.getSuper() != null);


        return null;
    }
}
/*
 * 
 * class Foo{
 * 
 * private String str;
 * private int a ;
 * private Bar bar;
 * 
 * public Foo(){
 * 
 *  bar.str = "Hallo";
 *  a = 10;
 * }
 * 
 * }
 * 
 * 
 * 
 * 
 * 
 */
