/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package symbol;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import teajay.bytecode.attributes.code.FullFrame;
import teajay.bytecode.attributes.code.JVMInstruction;
import teajay.bytecode.attributes.code.StackMapTable;
import teajay.bytecode.attributes.code.VerificationType;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.exceptions.MalformedClassNameException;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
class StackMapGenerator {

    private enum StackMapType {

        Same, Append, Chop1, Chop2, Chop3, Full;
    }
    private final LocalTable localTable;
    private final StackMapTable table;
    private List<VerificationType> lastFrame;

    StackMapGenerator(LocalTable localTable, StackMapTable table) {
        this.localTable = localTable;
        this.table = table;
        this.lastFrame = table.getFirstFrame().getLocalTypeList().getList();
    }

    void addStackMapFrame(JVMInstruction start, List<VerificationType> stack) {
        try {
            List<VerificationType> currentFrame = new ArrayList<>();
            for (Frame frame : localTable) {
                for (LocalEntry entry : frame) {
                    currentFrame.add(VerificationType.valueOf(new ClassInfo(entry.getQID())));
                }
            }
            switch (stack.size()) {
                case 0: {
                    switch (calcType(lastFrame, currentFrame)) {
                        case Same:
                            table.addSameFrame(start);
                            break;
                        case Append:
                            table.addAppendFrame(start).addLocal(currentFrame.subList(lastFrame.size(), currentFrame.size()));
                            break;
                        case Chop1:
                            table.addChopFrame(start, 1);
                            break;
                        case Chop2:
                            table.addChopFrame(start, 2);
                            break;
                        case Chop3:
                            table.addChopFrame(start, 3);
                            break;
                        default:
                            table.addFullFrame(start).addLocal(currentFrame);
                    }
                    break;
                }
                case 1: {
                    switch (calcType(lastFrame, currentFrame)) {
                        case Same:
                            table.addSameLocalsOneStackItemFrame(start).addStack(stack);
                            break;
                        default:
                            FullFrame full = table.addFullFrame(start);
                            full.addLocal(currentFrame);
                            full.addStack(stack);
                    }
                    break;
                }
                default: {
                    FullFrame full = table.addFullFrame(start);
                    full.addLocal(currentFrame);
                    full.addStack(stack);
                }
            }
            this.lastFrame = currentFrame;
        } catch (ByteCodeException ex) {
            throw new RuntimeException(ex);
        }
    }

    static StackMapType calcType(List<VerificationType> last, List<VerificationType> current) {
        if (last.size() < current.size()) {
            int i;
            for (i = 0; i < last.size(); i++) {
                if (!last.get(i).equals(current.get(i))) {
                    return StackMapType.Full;
                }
            }
            if (current.size() - i <= 3) {
                return StackMapType.Append;
            }
        } else if (last.size() > current.size()) {
            int i;
            for (i = 0; i < current.size(); i++) {
                if (!last.get(i).equals(current.get(i))) {
                    return StackMapType.Full;
                }
            }
            switch (last.size() - i) {
                case 1:
                    return StackMapType.Chop1;
                case 2:
                    return StackMapType.Chop2;
                case 3:
                    return StackMapType.Chop3;
            }
        } else if (last.equals(current)) {
            return StackMapType.Same;
        }
        return StackMapType.Full;
    }

    public static void main(String[] args) throws MalformedClassNameException {
        VerificationType[] a = {VerificationType.makeObject(new ClassInfo("woot")), VerificationType.makeFloat(), VerificationType.makeLong(), VerificationType.makeNull()};
        VerificationType[] b = {VerificationType.makeObject(new ClassInfo("woot")), VerificationType.makeFloat()};
        System.err.println(StackMapGenerator.calcType(Arrays.asList(a), Arrays.asList(b)));
    }
}
