/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.generics.check;

import java.io.IOException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;
import stage3.utils.InformationCenter;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.generics.parser.Type;
import teajay.runtime.exceptions.ResolveException;
import teajay.util.QIDTools;
import teajay.util.io.ClassFileRef;
import teajay.util.io.ClassFinder;
import teajay.util.tjattribute.TeaJayAttribute;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
class ImplCheckVisitor implements Type.Visitor<Void, Void> {

    private String excludedImplName = "";
    
    ImplCheckVisitor() {
    }
    
    void reset() {
        excludedImplName = "";
    }
    
    void init(String exludedImplName) {
        excludedImplName = exludedImplName;
    }

    @Override
    public Void visitSimpleType(Type.SimpleType type, Void p) {
        return null;
    }

    @Override
    public Void visitArrayType(Type.ArrayType type, Void p) {
        return type.elemType.accept(this, null);
    }

    @Override
    public Void visitMethodType(Type.MethodType type, Void p) {
        if (type.paramTypes != null) {
            for (Type t : type.paramTypes) {
                t.accept(this, null);
            }
        }
        if (type.returnType != null) {
            type.returnType.accept(this, null);
        }
        if (type.throwsTypes != null) {
            for (Type t : type.throwsTypes) {
                t.accept(this, null);
            }
        }
        if (type.typeParamTypes != null) {
            for (Type.TypeParamType t : type.typeParamTypes) {
                t.accept(this, null);
            }
        }
        return null;
    }

    @Override
    public Void visitClassSigType(Type.ClassSigType type, Void p) {
        if (type.superclassType != null) {
            type.accept(this, null);
        }
        if (type.superinterfaceTypes != null) {
            for (Type t : type.superinterfaceTypes) {
                t.accept(this, null);
            }
        }
        if (type.typeParamTypes != null) {
            for (Type.TypeParamType t : type.typeParamTypes) {
                t.accept(this, null);
            }
        }
        return null;
    }

    @Override
    public Void visitClassType(Type.ClassType type, Void p) {
        try {
            String qid = type.name;
            ClassFileRef find = ClassFinder.instance().find(qid);
            if (find == null) {
                throw new TypeException("Class not found: " + qid);
            }
            TeaJayAttribute.Type teaJayAttributeType = InformationCenter.getTeaJayAttributeType(find.get());

            if (teaJayAttributeType == TeaJayAttribute.Type.Impl && !excludedImplName.equals(qid)) {
                throw new TypeException("Cannot use implementations outside their own namespace");
            } else if (find.get().isPackagePrivate()) {
                if (!QIDTools.samePackage(QIDTools.packageName(qid), QIDTools.packageName(excludedImplName))) {
                    throw new TypeException(qid+ " is package private");
                }
            }
        } catch (ByteCodeException | IOException | ResolveException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
        if (type.outerType != null) {
            type.outerType.accept(this, null);
        }
        if (type.typeArgs != null) {
            for (Type t : type.typeArgs) {
                t.accept(this, null);
            }
        }
        return null;
    }

    @Override
    public Void visitTypeParamType(Type.TypeParamType type, Void p) {
        if (type.classBound != null) {
            type.classBound.accept(this, null);
        }
        if (type.interfaceBounds != null) {
            for (Type t : type.interfaceBounds) {
                t.accept(this, null);
            }
        }
        return null;
    }

    @Override
    public Void visitWildcardType(Type.WildcardType type, Void p) {
        if (type.boundType != null) {
            type.boundType.accept(this, null);
        }
        return null;
    }
}
