/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.lang.impl;

import java.util.Arrays;
import java.util.Objects;
import teajay.lang.List;
import teajay.lang.Number;
import teajay.lang.TJIterable;
import teajay.lang.TJIterator;
import teajay.lang.TJObject;
import teajay.util.TJObjects;
import teajay.util.TJType;
import teajay.util.tjattribute.TeaJayAttribute;

/**
 * Default List implementation.
 *
 * @author Michael Kußmann <michi@outx.de>
 * @param <T>
 */
@TJType(type = TeaJayAttribute.Type.Impl)
public final class ListImpl<T> extends List<T> {

    private static final Number DEFAULT_CAPACITY = Number.$valueOf(16);
    private final List<T> rthis;
    private T[] array;
    private Number size = Number.ZERO;
    private final TJObject rsuper;

    public ListImpl(List<T> rthis) {
        this(rthis, DEFAULT_CAPACITY);
    }

    public ListImpl(List<T> rthis, T... values) {
        this(rthis, values, true);
    }

    public ListImpl(List<T> rthis, T head, List<? extends T> tail) {
        this(rthis, tail.size().add(Number.ONE));
        this.rthis.add(head);
        this.rthis.addAll(tail);
    }

    private static Number size(List<?>... lsts) {
        Number size = Number.ZERO;
        for (List<?> lst : lsts) {
            size = size.add(lst.size());
        }
        return size;
    }

    public ListImpl(List<T> rthis, List<? extends T>... lists) {
        this(rthis, size(lists));
        for (List<? extends T> lst : lists) {
            this.rthis.addAll(lst);
        }
    }

    @SuppressWarnings("unchecked")
    private ListImpl(List<T> rthis, Number initialCapacity) {
        if (!initialCapacity.isInteger() || initialCapacity.isNegative()) {
            throw new Error("Dimension value of the array must be an integer greater than or equal to 0");
        }
        this.array = (T[]) new Object[initialCapacity.$int_val()];
        if (rthis != null) {
            this.rthis = rthis;
        } else {
            this.rthis = this;
        }
        this.rsuper = new TJObject(this.rthis);
    }

    @SuppressWarnings("unchecked")
    private ListImpl(List<T> rthis, T[] data, boolean copy) {
        if (data.getClass().getComponentType() != null && data.getClass().getComponentType().isArray()) {
            throw new Error("Can only create teajay lists from 1D java arrays");
        }
        if (copy) {
            this.array = (T[]) new Object[data.length];
            System.arraycopy(data, 0, array, 0, data.length);
        } else {
            this.array = data;
        }
        this.size = Number.$valueOf(data.length);
        if (rthis != null) {
            this.rthis = rthis;
        } else {
            this.rthis = this;
        }
        this.rsuper = new TJObject(this.rthis);
    }

    private void ensureCapacity(Number nindex) {
        int index = nindex.$int_val();
        if (index >= array.length) {
            int newSize;
            if (array.length <= 1024) {
                newSize = array.length * 3;
            } else {
                newSize = array.length * 2;
            }
            if (index >= newSize) {
                newSize = index + 1;
            }
            T[] newArray = (T[]) new Object[newSize];
            System.arraycopy(array, 0, newArray, 0, array.length);
            this.array = newArray;
        }
    }

    private void rangeCheck(Number index) {
        if (index.isNegative() || !index.isInteger() || index.compareTo(size) >= 0) {
            throw new IndexOutOfBoundsException("List index out of bounds: " + index);
        }
    }

    @Override
    public Boolean isEmpty() {
        return size.equals(Number.ZERO);
    }

    @Override
    public T operator_get(Number index) {
        return rthis.get(index);
    }

    @Override
    public void operator_set(Number index, T value) {
        rthis.set(index, value);
    }

    @Override
    public T get(Number index) throws IndexOutOfBoundsException {
        rangeCheck(index);

        return array[index.$int_val()];
    }

    @Override
    public void set(Number index, T value) throws IndexOutOfBoundsException {
        if (index.isNegative() || !index.isInteger()) {
            throw new IndexOutOfBoundsException("List index out of bounds: " + index);
        }
        ensureCapacity(index);
        array[index.$int_val()] = value;
        if (index.compareTo(size) >= 0) {
            this.size = index.add(Number.ONE);
        }
    }

    @Override
    public void add(T value) {
        rthis.set(size, value);
    }

    @Override
    public void addAll(TJIterable<? extends T> vals) {
        TJIterator<? extends T> it = vals.iterator();
        while (it.hasNext()) {
            rthis.add(it.next());
        }
    }

    @Override
    public void insert(Number index, T value) throws IndexOutOfBoundsException {
        if (index.isNegative() || !index.isInteger()) {
            throw new IndexOutOfBoundsException("List index out of bounds: " + index);
        }
        if (index.compareTo(size) >= 0) {
            rthis.set(index, value);
        } else {
            ensureCapacity(size);
            int isize = size.$int_val();
            int iindex = index.$int_val();
            int numMoved = isize - iindex;
            System.arraycopy(array, iindex, array, iindex + 1, numMoved);
            array[iindex] = value;
            this.size = size.add(Number.ONE);
        }
    }

    @Override
    public T remove(Number index) {
        rangeCheck(index);

        T oldValue = rthis.operator_get(index);
        int isize = size.$int_val();
        int iindex = index.$int_val();
        int numMoved = isize - iindex - 1;
        if (numMoved > 0) {
            System.arraycopy(array, iindex + 1, array, iindex, numMoved);
        }
        array[isize - 1] = null;
        this.size = size.sub(Number.ONE);
        return oldValue;
    }

    @Override
    public Number indexOf(T obj) {
        int index = 0;
        TJIterator<?> it = rthis.iterator();
        while (it.hasNext()) {
            Object next = it.next();
            if (next == obj) {
                return Number.$valueOf(index);
            } else if (next != null && TJObjects.equals(next, obj)) {
                return Number.$valueOf(index);
            }
            index++;
        }
        return Number.M_ONE;
    }

    private boolean removeFirst(T elem) {
        Number index = rthis.indexOf(elem);
        if (index.isNegative()) {
            return false;
        }
        rthis.remove(index);
        return true;
    }

    @Override
    public Boolean removeAll(T elem) {
        Boolean found = removeFirst(elem);
        while (removeFirst(elem)) {
            ;
        }
        return found;
    }

    @Override
    public List<T> subList(Number fromIndex) {
        return rthis.subList(fromIndex, size);
    }

    @Override
    public List<T> subList(Number fromIndex, Number endIndex) {
        int from = fromIndex.$int_val();
        int end = endIndex.$int_val();
        int size = this.size.$int_val();
        if (from == size && end == size) {
            return new ListImpl<>(null);
        }
        rangeCheck(fromIndex);
        rangeCheck(endIndex.sub(Number.ONE));
        return new ListImpl<>(null, Arrays.copyOfRange(array, from, end), false);
    }

    @Override
    public Number size() {
        return size;
    }

    @Override
    public List<T> clone() {
        return new ListImpl<>(null, Arrays.copyOf(array, size.$int_val()), false);
    }

    @Override
    public String toString() {
        StringBuilder buf = new StringBuilder();
        TJIterator<?> it = rthis.iterator();
        buf.append('[');
        if (it.hasNext()) {
            buf.append(String.valueOf(it.next()));
        }
        while (it.hasNext()) {
            buf.append(", ").append(String.valueOf(it.next()));
        }
        buf.append(']');
        return buf.toString();
    }

    @Override
    public TJIterator<T> iterator() {
        return new TJIterator<T>() {
            private Number pos = Number.ZERO;

            @Override
            public Boolean hasNext() {
                return pos.compareTo(rthis.size()) < 0;
            }

            @Override
            public T next() {
                try {
                    T tmp = rthis.get(pos);
                    pos = pos.add(Number.ONE);
                    return tmp;
                } catch (IndexOutOfBoundsException ex) {
                    throw new IllegalStateException("no such element");
                }
            }

            @Override
            public void remove() {
                if (pos.equals(Number.ZERO)) {
                    throw new IllegalStateException("no such element");
                }
                pos = pos.sub(Number.ONE);
                rthis.remove(pos);
            }
        };
    }

    @Override
    public Boolean operator_eq(Object obj) {
        if (obj instanceof List) {
            List<?> list = (List<?>) obj;
            if (list.size().equals(rthis.size())) {
                TJIterator<?> ita = list.iterator();
                TJIterator<?> itb = rthis.iterator();
                while (itb.hasNext()) {
                    Object a = ita.next();
                    Object b = itb.next();
                    if (a == null || b == null) {
                        if (a != b) {
                            return false;
                        } //else: a and b are both null and thus equal
                    } else if (!a.equals(b)) {
                        return false;
                    }
                }
            } else {
                return false;
            }
            return true;
        } else {
            return false;
        }
    }

    @Override
    public Number tjHashCode() {
        int hashCode = 7;
        TJIterator<?> it = rthis.iterator();
        while (it.hasNext()) {
            hashCode *= Objects.hashCode(it.next()) + 3;
        }
        return Number.$valueOf(hashCode);
    }

    @Override
    public Boolean contains(T elem) {
        return rthis.indexOf(elem).compareTo(Number.ZERO) >= 0;
    }

    @Override
    public void clear() {
        this.size = Number.ZERO;
        Arrays.fill(array, null);
    }

    /**
     * This method is not part of the public interface of teajay.lang.List and
     * only for internal optimizations.
     *
     * @return a reference to the array holding the list's elements
     */
    public T[] getArray() {
        return array;
    }

    @Override
    public int hashCode() {
        return rsuper.hashCode();
    }

    @Override
    public boolean equals(Object obj) {
        return rsuper.equals(obj);
    }
}