/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.runtime;

import java.lang.invoke.*;
import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import teajay.runtime.exceptions.TeaJayResolveError;
import teajay.runtime.exceptions.TeaJayRuntimeError;
import teajay.util.TJType;
import teajay.util.tjattribute.TeaJayAttribute;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
@TJType(type = TeaJayAttribute.Type.Hidden)
public final class Bootstraps {

    private static final TypeManager tm = TypeManager.instance();
    private static Map<Class<?>, Map<MethodType, CallSite>> constructorCallSiteCache;
    private static Map<Class<?>, Map<String, Map<MethodType, CallSite>>> staticCallSiteCache;
    private static Map<CallSite, MethodHandles.Lookup> lookups;
    /**
     * Very experimental feature.
     */
    private volatile static boolean ALLOW_TYPE_RESET = false;
    private volatile static boolean used = false;

    synchronized static void enableTypereset() {
        if (!ALLOW_TYPE_RESET) {
            if (used) {
                throw new IllegalStateException("There have already been bootstrap operations.");
            }
            ALLOW_TYPE_RESET = true;
            constructorCallSiteCache = new HashMap<>(1024);
            staticCallSiteCache = new HashMap<>(1024);
            lookups = new HashMap<>(2048);
        }
    }

    synchronized static boolean isTyperesetEnabled() {
        return ALLOW_TYPE_RESET;
    }

    synchronized static void resetType(Class<?> type) throws NoSuchMethodException, IllegalAccessException {
        if (!ALLOW_TYPE_RESET) {
            throw new TeaJayRuntimeError("Switching types at runtime is disabled.");
        }
        Class<?> impl = tm.getImpl(type);

        Map<MethodType, CallSite> map = constructorCallSiteCache.get(type);
        if (map != null) {
            for (Entry<MethodType, CallSite> entry : map.entrySet()) {
                CallSite callSite = entry.getValue();

                MethodType mtype = entry.getKey();
                MethodType voidType = mtype.changeReturnType(void.class);
                MethodHandles.Lookup callerClass = lookups.get(callSite);
                if (callerClass == null) {
                    throw new Error("Fatal runtime error while switching types: callerClass == null");
                }
                callSite.setTarget(callerClass.findConstructor(impl, voidType).asType(mtype));
            }
        }
        Map<String, Map<MethodType, CallSite>> map2 = staticCallSiteCache.get(type);
        if (map2 != null) {
            for (Entry<String, Map<MethodType, CallSite>> tmp : map2.entrySet()) {
                for (Entry<MethodType, CallSite> entry : tmp.getValue().entrySet()) {
                    CallSite callSite = entry.getValue();
                    MethodHandles.Lookup callerClass = lookups.get(callSite);
                    if (callerClass == null) {
                        throw new Error("Fatal runtime error while switching types: callerClass == null");
                    }
                    String name = tmp.getKey();
                    MethodType mtype = entry.getKey();
                    callSite.setTarget(callerClass.findStatic(impl, name, mtype));
                }
            }
        }
    }
    /**
     * Creates a new TeaJay-Type.
     * @param callerClass
     * @param name
     * @param mtype
     * @param target
     * @return
     * @throws Exception
     * @throws TeaJayResolveError 
     */
    public static CallSite bc(MethodHandles.Lookup callerClass, String name, MethodType mtype, Class<?> target) throws Exception, TeaJayResolveError {
        used = true; //Race conditions possible, but this is only to protect, at best effort, against incorrect usage
        int mod = target.getModifiers();
        MethodType voidType = mtype.changeReturnType(void.class);
        Class<?> impl = target;
        if (Modifier.isAbstract(mod) || Modifier.isInterface(mod)) {
            //Possible TeaJay type
            impl = tm.getImpl(target);
            if (ALLOW_TYPE_RESET) {
                CallSite callSite;
                synchronized (Bootstraps.class) {
                    Map<MethodType, CallSite> tmp = constructorCallSiteCache.get(target);
                    if (tmp == null) {
                        tmp = new HashMap<>();
                        constructorCallSiteCache.put(target, tmp);
                    }
                    callSite = tmp.get(mtype);
                    if (callSite == null) {
                        callSite = new MutableCallSite(callerClass.findConstructor(impl, voidType).asType(mtype));
                        tmp.put(mtype, callSite);
                        lookups.put(callSite, callerClass);
                    }
                }
                return callSite;
            }
        }
        return new ConstantCallSite(callerClass.findConstructor(impl, voidType).asType(mtype));
    }
    /**
     * Calls a static Method.
     * @param callerClass
     * @param name
     * @param mtype
     * @param target
     * @return
     * @throws Exception
     * @throws TeaJayResolveError 
     */
    public static CallSite st(MethodHandles.Lookup callerClass, String name, MethodType mtype, Class<?> target) throws Exception, TeaJayResolveError {
        used = true;
        int mod = target.getModifiers();
        Class<?> impl = target;
        if (Modifier.isAbstract(mod) || Modifier.isInterface(mod)) {
            impl = tm.getImpl(target);
            if (ALLOW_TYPE_RESET) {
                CallSite callSite;
                synchronized (Bootstraps.class) {
                    Map<String, Map<MethodType, CallSite>> tmp = staticCallSiteCache.get(target);
                    if (tmp == null) {
                        tmp = new HashMap<>();
                        staticCallSiteCache.put(target, tmp);
                    }
                    Map<MethodType, CallSite> tmp2 = tmp.get(name);
                    if (tmp2 == null) {
                        tmp2 = new HashMap<>();
                        tmp.put(name, tmp2);
                    }
                    callSite = tmp2.get(mtype);
                    if (callSite == null) {
                        callSite = new MutableCallSite(callerClass.findStatic(impl, name, mtype));
                        tmp2.put(mtype, callSite);
                        lookups.put(callSite, callerClass);
                    }
                    return callSite;
                }
            }
        }
        return new ConstantCallSite(callerClass.findStatic(impl, name, mtype));
    }
}