/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.runtime;

import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;

/**
 * The default type resolver.
 * @author Michael Kußmann <michi@outx.de>
 */
final class PromptImplementationFinder implements ImplementationFinder {

    private final Reader rd = new InputStreamReader(System.in);

    PromptImplementationFinder() {
    }

    @Override
    public <T> Class<? extends T> findImplementation(Class<T> type) {
        String implName;
        try {
            System.out.print("Please state an implementation for " + type.getName() + ": ");
            System.out.flush();
            implName = readLine(rd);
            if (implName != null && !implName.equals("")) {
                Class<? extends T> impl = (Class<? extends T>) PromptImplementationFinder.class.getClassLoader().loadClass(implName);
                if (type.isAssignableFrom(impl)) {
                    return impl;
                } else {
                    System.err.println("The given implementation does not implement " + type.getName() + ".");
                }
            } else {
                return null;
            }
        } catch (IOException ex) {
            System.err.println(ex.getLocalizedMessage());
            return null;
        } catch (ClassNotFoundException ex) {
            System.err.println("Class not found: " + ex.getMessage());
        }
        return findImplementation(type);
    }
    /**
     * Read a line without reading ahead like a BufferedReader would do.
     * @return
     * @throws IOException 
     */
    private static String readLine(Reader rd) throws IOException {
        StringBuilder buf = new StringBuilder();
        char[] c  = new char[1];
        while (rd.read(c) > 0 && c[0] != '\n') {
            if (c[0] != '\r') {
                buf.append(c[0]);
            }
        }
        return buf.toString();
    }
}
