/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.runtime;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Modifier;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import teajay.runtime.exceptions.TeaJayResolveError;
import teajay.runtime.exceptions.TeaJayRuntimeError;
import teajay.util.TJType;
import teajay.util.io.FileIterator;
import teajay.util.io.FilesVisitor;
import teajay.util.io.Input;
import teajay.util.io.StopScan;
import teajay.util.tjattribute.TeaJayAttribute;

/**
 * Managment of TeaJay types and their implementations.
 *
 * @author Michael Kußmann <michi@outx.de>
 */
@TJType(accessType = TeaJayAttribute.AccessType.Abstract)
public final class TypeManager {

    public static final String runtimePath = "teajay.runtime.path";
    public static final String mainConfig = "teajay.main.config";
    private volatile boolean warn = Boolean.getBoolean("tj_warn");
    private ClassLoader jarLoader;

    private class TypeFinder implements FilesVisitor {

        @Override
        public void visit(String javaPackage, String filename, Input in) throws StopScan {
            putTypes(javaPackage, in, false);
        }

        @Override
        public void failed(File file, IOException reason) throws StopScan {
            throw new TeaJayRuntimeError("Corrupted teajay runtime path: " + reason.getLocalizedMessage(), reason);
        }

        @Override
        public boolean accept(File file) {
            String name = file.getName().toLowerCase();
            return name.endsWith(".tjt");
        }
    }
    private static TypeManager instance;
    private static final String pathSeparator = System.getProperty("path.separator", ":");
    private ClassLoader loader;
    private final Map<Class<?>, Object> types = new HashMap<>(1024);
    private volatile ImplementationFinder implFinder = null;

    private TypeManager() throws TeaJayRuntimeError {
        List<URL> urls = new ArrayList<>();
        File[] teajayRuntimePath;
        String rp = System.getProperty(runtimePath);
        boolean hasRP = true;
        String jarFile = System.getProperty("tj_jar");
        if (rp == null) {
            hasRP = false;
            if (jarFile == null) {
                rp = "." + pathSeparator + System.getProperty("java.class.path");
            } else {
                rp = System.getProperty("java.class.path") + pathSeparator + jarFile;
            }
        } else {
            if (jarFile == null) {
                rp += pathSeparator + System.getProperty("java.class.path");
            } else {
                rp += pathSeparator + System.getProperty("java.class.path") + pathSeparator + jarFile;
            }
        }
        try {
            String[] paths = rp.split(pathSeparator);
            if (paths.length <= 0) {
                throw new TeaJayRuntimeError("Cannot determine the teajay runtime path.");
            }
            List<File> files = new ArrayList<>();

            for (int i = 0; i < paths.length; i++) {
                File file = new File(paths[i]);
                if (!file.canRead()) {
                    throw new TeaJayRuntimeError("Corrupted teajay runtime path or class path: " + file);
                } else {
                    if (!hasRP || !file.getName().equals(".")) {
                        files.add(file);
                    }
                    urls.add(file.toURI().toURL());
                }
            }
            teajayRuntimePath = files.toArray(new File[0]);
            if (hasRP) {
                this.loader = new URLClassLoader(urls.toArray(new URL[0]), TypeManager.class.getClassLoader());
            } else {
                if (jarFile == null) {
                    this.loader = new URLClassLoader(new URL[]{new File(".").toURI().toURL()}, TypeManager.class.getClassLoader());
                } else {
                    this.loader = new URLClassLoader(new URL[]{new File(jarFile).toURI().toURL()}, TypeManager.class.getClassLoader());
                }
            }
            FileIterator it = new FileIterator(teajayRuntimePath);
            it.scan(new TypeFinder(), true);

            String maincfg = System.getProperty(mainConfig);
            final InputStream ressource;
            if (maincfg == null) {
                ressource = getClass().getResourceAsStream("/teajay.conf");
            } else {
                try {
                    ressource = new FileInputStream(maincfg);
                } catch (IOException ex) {
                    throw new TeaJayRuntimeError("failed to read teajay config: " + ex.getLocalizedMessage());
                }
            }
            if (ressource != null) {

                putTypes(null, new Input() {
                    @Override
                    public InputStream openStream() throws IOException {
                        return ressource;
                    }

                    @Override
                    public File getFile() {
                        return null;
                    }
                }, true);
            }
            try {
                it.close();
            } catch (IOException ex) {
                //ex.printStackTrace(System.err);
            }
        } catch (MalformedURLException ex) {
            throw new Error("fatal error", ex);
        }
    }

    private static String getPackage(String qID) {
        int index = qID.lastIndexOf(".");
        if (index >= 0) {
            return qID.substring(0, index);
        }
        return null;
    }

    /**
     * Returns the class loader that has been initialized with TeaJay's runtime
     * path and class path.
     *
     * @return the class loader
     */
    public ClassLoader getClassLoader() {
        return this.loader;
    }
    
    public ClassLoader getJarLoader() {
        return this.jarLoader;
    }

    void putTypes(String javaPackage, Input input, boolean override) {
        Properties p = new Properties();
        try (InputStream in = input.openStream()) {
            p.load(in);
            putTypes(javaPackage, p, input.toString(), override);
        } catch (IOException ex) {
            if (warn) {
                System.err.println("[WARN] corrupted type definition file: " + ex.getLocalizedMessage());
            }
        }
    }

    void putTypes(String javaPackage, Properties p, String source, boolean override) {
        if (!override && p.size() > 1) {
            throw new TeaJayRuntimeError(".tjt files may not state more than one Type - Implementation pair: " + source);
        }
        for (Entry<Object, Object> entry : p.entrySet()) {
            String typeName = entry.getKey().toString();
            if (typeName.contains("/")) {
                throw new TeaJayRuntimeError("invalid character / in: " + source);
            }
            String externalPackageName;
            if (javaPackage != null) {
                externalPackageName = javaPackage.replace('/', '.') + ".";
            } else {
                externalPackageName = "";
            }
            try {
                Class<?> type = loader.loadClass(typeName);
                if (types.containsKey(type)) {
                    /*if (typeName.startsWith("teajay/lang") && !typeName.equals("teajay/lang/List")) {
                        throw new TeaJayRuntimeError("Cannot override implementations for teajay.lang.* types (except for List)");
                    }*/
                    if (!override && warn) {
                        System.err.println("[WARN] found duplicate implementation for " + typeName);
                    }
                }
                String value = entry.getValue().toString();
                if (value.contains("/")) {
                    throw new TeaJayRuntimeError("invalid character / in: " + source);
                }
                String impl;
                if (!override && getPackage(value) != null) {
                    throw new TeaJayRuntimeError(".tjt files may not state a package name for the implementation part: " + source);
                } else {
                    impl = externalPackageName + value;
                }
                types.put(type, impl);
            } catch (ClassNotFoundException ex) {
                if (warn || override) {
                    System.err.println("[WARN] no such type " + typeName);
                }
            }
        }
    }

    private <T> Class<? extends T> findImplementation(Class<T> cl) {
        if (implFinder != null) {
            return implFinder.findImplementation(cl);
        }
        return null;
    }

    /**
     * Singleton.
     *
     * @return the one and only instance
     */
    public synchronized static TypeManager instance() {
        if (instance == null) {
            instance = new TypeManager();
        }
        return instance;
    }

    /**
     * Sets a user defined implementation finder.
     *
     * @param implFinder
     */
    public void setImplementationFinder(ImplementationFinder implFinder) {
        this.implFinder = implFinder;
    }

    synchronized <T> Class<? extends T> getImpl(Class<T> cl) throws TeaJayResolveError {
        try {
            Object obj = types.get(cl);
            Class<? extends T> impl;
            if (obj == null || obj instanceof String) {
                String name = (String) obj;
                if (name == null) {
                    impl = findImplementation(cl);
                } else {
                    impl = (Class<? extends T>) loader.loadClass(name);
                }
                if (impl == null) {
                    throw new TeaJayResolveError("Cannot find an implementation for " + cl.getName());
                } else if (Modifier.isAbstract(impl.getModifiers())) {
                    throw new TeaJayRuntimeError("Implementation " + impl.getName() + " for " + cl.getName() + " may not be abstract.");
                } else if (!cl.isAssignableFrom(impl)) {
                    throw new TeaJayRuntimeError(impl.getName() + " does not implement " + cl.getName());
                }
                types.put(cl, impl);
            } else if (obj instanceof Class<?>) {
                impl = (Class<? extends T>) obj;
            } else {
                throw new Error("Fatal runtime error, type map is corrupted with: " + obj.getClass().getName());
            }
            return impl;
        } catch (ClassNotFoundException ex) {
            throw new TeaJayResolveError("Cannot find an implementation for " + cl.getName(), ex);
        }
    }

    /**
     * Do not use.
     *
     * @param type
     * @param impl
     * @throws TeaJayRuntimeError
     */
    public synchronized void setTypeImpl(Class<?> type, String impl) throws TeaJayRuntimeError {
        if (!Bootstraps.isTyperesetEnabled()) {
            throw new TeaJayRuntimeError("Switching types at runtime is disabled.");
        }
        try {
            Class<?> implClass = loader.loadClass(impl);
            if (type.isAssignableFrom(implClass)) {
                types.put(type, implClass);
                Bootstraps.resetType(type);
            } else {
                throw new TeaJayResolveError("Implementations must be assignable to their type definition");
            }
        } catch (NoSuchMethodException | IllegalAccessException ex) {
            throw new TeaJayResolveError("Invalid implementation", ex);
        } catch (ClassNotFoundException ex) {
            throw new TeaJayResolveError("Class not found " + impl, ex);
        }
    }
    
    public void enableWarnings() {
        this.warn = true;
    }
    
    public void disableWarnings() {
        this.warn = false;
    }

    public Boolean canSwitchTypes() {
        return Bootstraps.isTyperesetEnabled();
    }
}