/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.util;

import teajay.util.tjattribute.TeaJayAttribute;
import teajay.lang.Number;
import teajay.lang.TJComparable;
import teajay.lang.TJObject;

/**
 * Fast numeric calculations with double precision.
 *
 * @author Michael Kußmann <michi@outx.de>
 */
@TJType(accessType = TeaJayAttribute.AccessType.Final)
public final class Real extends TJObject implements Comparable<Real>, TJComparable<Real> {

    private final double value;

    public Real(Number num) {
        value = num.$double_val();
    }

    private Real(double value) {
        this.value = value;
    }

    public double doubleValue() {
        return value;
    }

    public float floatValue() {
        return (float) value;
    }

    public int intValue() {
        return (int) value;
    }

    public long longValue() {
        return (long) value;
    }

    public static Real valueOf(double val) {
        return new Real(val);
    }

    public static Real valueOf(float val) {
        return new Real(val);
    }

    public static Real valueOf(long val) {
        return new Real(val);
    }

    public static Real valueOf(int val) {
        return new Real(val);
    }

    public static Real valueOf(Number val) {
        return new Real(val);
    }

    public static Real operator_add(Real a, Real b) {
        return new Real(a.value + b.value);
    }

    public static Real operator_sub(Real a, Real b) {
        return new Real(a.value - b.value);
    }

    public static Real operator_mult(Real a, Real b) {
        return new Real(a.value * b.value);
    }

    public static Real operator_div(Real a, Real b) {
        return new Real(a.value / b.value);
    }

    public static Real operator_add(Number a, Real b) {
        return new Real(a.$double_val() + b.value);
    }

    public static Real operator_add(Real a, Number b) {
        return new Real(a.value + b.$double_val());
    }

    public static Real operator_sub(Number a, Real b) {
        return new Real(a.$double_val() - b.value);
    }

    public static Real operator_sub(Real a, Number b) {
        return new Real(a.value - b.$double_val());
    }

    public static Real operator_sub(Real a) {
        return new Real(-a.value);
    }

    public static Real operator_mult(Number a, Real b) {
        return new Real(a.$double_val() * b.value);
    }

    public static Real operator_mult(Real a, Number b) {
        return new Real(a.value * b.$double_val());
    }

    public static Real operator_div(Number a, Real b) {
        return new Real(a.$double_val() / b.value);
    }

    public static Real operator_div(Real a, Number b) {
        return new Real(a.value / b.$double_val());
    }

    @Override
    public Boolean operator_eq(Object other) {
        return other instanceof Real && ((Real) other).value == value;
    }

    public Real pow(Real exp) {
        return new Real(java.lang.Math.pow(value, exp.value));
    }

    public Real pow(Number exp) {
        return new Real(java.lang.Math.pow(value, exp.$double_val()));
    }

    public Real sqrt() {
        return new Real(java.lang.Math.sqrt(value));
    }

    public Real log() {
        return new Real(java.lang.Math.log(value));
    }

    public Real log10() {
        return new Real(java.lang.Math.log10(value));
    }
    
    public Real abs() {
        return new Real(java.lang.Math.abs(value));
    }

    public Number toNumber() {
        return Number.$valueOf(value);
    }

    public Boolean isNaN() {
        return Double.isNaN(value);
    }

    public Boolean isInfinite() {
        return Double.isInfinite(value);
    }
    
    public static Boolean operator_less(Real a, Real b) {
        return a.value < b.value;
    }

    public static Boolean operator_less(Number a, Real b) {
        return a.$double_val() < b.value;
    }

    public static Boolean operator_less(Real a, Number b) {
        return a.value < b.$double_val();
    }

    public static Boolean operator_greater(Real a, Real b) {
        return a.value > b.value;
    }

    public static Boolean operator_greater(Number a, Real b) {
        return a.$double_val() > b.value;
    }

    public static Boolean operator_greater(Real a, Number b) {
        return a.value > b.$double_val();
    }

    public static Boolean operator_geq(Real a, Real b) {
        return a.value >= b.value;
    }

    public static Boolean operator_geq(Number a, Real b) {
        return a.$double_val() >= b.value;
    }

    public static Boolean operator_geq(Real a, Number b) {
        return a.value >= b.$double_val();
    }

    public static Boolean operator_leq(Real a, Real b) {
        return a.value <= b.value;
    }

    public static Boolean operator_leq(Number a, Real b) {
        return a.$double_val() <= b.value;
    }

    public static Boolean operator_leq(Real a, Number b) {
        return a.value <= b.$double_val();
    }

    @Override
    public int compareTo(Real o) {
        if (value < o.value) {
            return -1;
        } else if (value > o.value) {
            return 1;
        } else {
            return 0;
        }
    }

    @Override
    public Number tjCompareTo(Real other) {
        return Number.$valueOf(this.compareTo(other));
    }
}
