/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.util.io;

import java.io.Closeable;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class FileIterator implements Closeable {

    private static final FileFilter dirFilter = new FileFilter() {
        @Override
        public boolean accept(File pathname) {
            return pathname.isDirectory();
        }
    };
    private static final String pathSeparator = System.getProperty("path.separator");
    private static final String fileSeparator = System.getProperty("file.separator");
    private static final boolean isNotUnixStyleFileSeparator = !fileSeparator.equals("/");
    public static final String CLASSPATH = System.getProperty("java.class.path");
    public static final String BOOTPATH = System.getProperty("sun.boot.class.path");
    private final boolean filesOnly;
    private final List<File> paths;
    private Set<File> scannedDirs = new HashSet<>();
    private final Collection<Closeable> streams = Collections.synchronizedList(new LinkedList<Closeable>());
    private Path currentPath = null;

    private static String makePackageName(String path) {
        if (path == null) {
            return null;
        }
        if (isNotUnixStyleFileSeparator) {
            path = path.replace(fileSeparator, "/");
        }
        if (path.endsWith("/")) {
            path = path.substring(0, path.length() - 1);
        }
        return path;
    }

    public FileIterator(File... paths) {
        this(false, paths);
    }

    public FileIterator(String... paths) {
        this(false, paths);
    }

    public FileIterator(List<File> paths) {
        this(false, paths);
    }

    public FileIterator(boolean filesOnly, File... paths) {
        this(filesOnly, Arrays.asList(paths));
    }

    public FileIterator(boolean filesOnly, String... paths) {
        File[] tmp = new File[paths.length];
        for (int i = 0; i < paths.length; i++) {
            tmp[i] = new File(paths[i]);
        }
        this.filesOnly = filesOnly;
        this.paths = Arrays.asList(tmp);
    }

    private FileIterator(boolean filesOnly, List<File> paths) {
        this.filesOnly = filesOnly;
        this.paths = paths;
    }

    public static FileIterator fromPathStrings(String... pathStrings) {
        return fromPathStrings(false, pathStrings);
    }

    public static FileIterator fromPathStrings(boolean filesOnly, String... pathStrings) {
        List<File> paths = new ArrayList<>();
        for (String path : pathStrings) {
            for (String s : path.split(pathSeparator)) {
                paths.add(new File(s));
            }
        }
        return new FileIterator(filesOnly, paths);
    }

    public void scan(FilesVisitor visitor, boolean scanJars) {
        try {
            for (File file : paths) {
                try {
                    this.currentPath = file.toPath();
                    if (filesOnly && file.isFile()) {
                        scan(file, visitor, scanJars);
                    } else if (!filesOnly) {
                        scan(file, visitor, scanJars);
                    }
                } catch (IOException ex) {
                    visitor.failed(file, ex);
                }
            }
        } catch (StopScan stop) {
        } finally {
            this.currentPath = null;
            this.scannedDirs = new HashSet<>();
        }
    }

    private void scan(final File file, FilesVisitor visitor, boolean scanJars) throws StopScan, IOException {
        String filename = file.getName();
        if (file.isDirectory()) {
            File abs = file.getAbsoluteFile();
            if (!scannedDirs.contains(abs)) {
                scannedDirs.add(abs);
                File[] files = file.listFiles(visitor);
                if (files != null) {
                    for (File f : files) {
                        scanFile(f, visitor);
                    }
                }
                files = file.listFiles(dirFilter);
                if (files != null) {
                    for (File f : files) {
                        try {
                            scan(f, visitor, false);
                        } catch (IOException ex) {
                            visitor.failed(file, ex);
                        }
                    }
                }
            }
        } else if (scanJars && filename.toLowerCase().endsWith(".jar")) {
            final ZipFile zipFile = new ZipFile(file);
            streams.add(zipFile);
            Enumeration<? extends ZipEntry> entries = zipFile.entries();

            while (entries.hasMoreElements()) {
                final ZipEntry entry = entries.nextElement();
                File f = new File(entry.getName());
                if (visitor.accept(f)) {
                    String parent = makePackageName(f.getParent());
                    visitor.visit(parent != null ? parent : null, f.getName(), new Input() {
                        @Override
                        public InputStream openStream() throws IOException {
                            return zipFile.getInputStream(entry);
                        }

                        @Override
                        public String toString() {
                            return file + "?" + entry.getName();
                        }

                        @Override
                        public File getFile() {
                            return null;
                        }
                    });
                }
            }
        } else if (visitor.accept(file)) {
            scanFile(file, visitor);
        }
    }

    private void scanFile(final File file, FilesVisitor visitor) throws StopScan {
        if (file.exists()) {
            Input input = new Input() {
                @Override
                public InputStream openStream() throws IOException {
                    return new FileInputStream(file);
                }

                @Override
                public String toString() {
                    return file.toString();
                }

                @Override
                public File getFile() {
                    return file;
                }
            };
            Path parent = currentPath.relativize(file.toPath()).getParent();

            visitor.visit(parent != null ? makePackageName(parent.toString()) : null, file.getName(), input);
        }
    }

    @Override
    public void close() throws IOException {
        IOException ex = null;
        synchronized (streams) {
            for (Closeable c : streams) {
                try {
                    c.close();
                } catch (IOException e) {
                    if (ex == null) {
                        ex = e;
                    }
                }
            }
        }
        if (ex != null) {
            throw ex;
        }
    }
}
