/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.util.tjattribute;

import java.io.File;
import java.io.IOException;
import java.util.List;
import teajay.bytecode.ClassFile;
import teajay.bytecode.attributes.annotations.Annotation;
import teajay.bytecode.attributes.annotations.ElementValue;
import teajay.bytecode.attributes.annotations.RuntimeInvisibleAnnotations;
import teajay.bytecode.attributes.signature.Signature;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.methods.MethodInfo;
import teajay.util.TJSignature;
import teajay.util.TJType;
import teajay.util.io.FileIterator;
import teajay.util.io.FilesVisitor;
import teajay.util.io.Input;
import teajay.util.io.StopScan;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class InjectTJAttribute {

    private static final FieldDescriptor overrideDescr = FieldDescriptor.fromClass(TJSignature.class);
    private static final FieldDescriptor tjtype = FieldDescriptor.fromClass(TJType.class);
    private static final TeaJayAttribute.AccessType defaultAccessType;
    private static final TeaJayAttribute.Type defaultType;

    static {
        TeaJayAttribute.AccessType atype;
        TeaJayAttribute.Type type;
        try {
            atype = (TeaJayAttribute.AccessType) TJType.class.getMethod("accessType").getDefaultValue();
            type = (TeaJayAttribute.Type) TJType.class.getMethod("type").getDefaultValue();
            if (atype == null) {
                atype = TeaJayAttribute.AccessType.None;
                System.err.println("[WARN] Unable to find default value for accessType");
            }
            if (type == null) {
                type = TeaJayAttribute.Type.TJBuiltinType;
                System.err.println("[WARN] Unable to find default value for type");
            }
        } catch (NoSuchMethodException ex) {
            System.err.println("[WARN] " + ex.getLocalizedMessage());
            atype = TeaJayAttribute.AccessType.None;
            type = TeaJayAttribute.Type.TJBuiltinType;
        }
        defaultAccessType = atype;
        defaultType = type;
    }

    private static void inject(File file) throws IOException, ByteCodeException {
        ClassFile classFile = ClassFile.read(file);
        RuntimeInvisibleAnnotations rans = (RuntimeInvisibleAnnotations) classFile.getAttribute(RuntimeInvisibleAnnotations.NAME);
        processSignatureOverride(rans, null, classFile);
        processTJType(rans, null, classFile);

        for (List<MethodInfo> methods : classFile.getMethods()) {
            for (MethodInfo info : methods) {
                rans = (RuntimeInvisibleAnnotations) info.getAttribute(RuntimeInvisibleAnnotations.NAME);
                processSignatureOverride(rans, info, null);
                processTJType(rans, info, null);
            }
        }
        classFile.write(file);
    }

    private static void processTJType(RuntimeInvisibleAnnotations annotations, MethodInfo info, ClassFile file) throws ByteCodeException {
        if (annotations != null) {
            for (Annotation a : annotations) {
                if (a.getType().equals(tjtype)) {
                    ElementValue typeValue = a.getElementValue("type");
                    ElementValue accessValue = a.getElementValue("accessType");
                    TeaJayAttribute.AccessType accessType;
                    TeaJayAttribute.Type type;
                    if (accessValue != null) {
                        accessType = accessValue.getValue(TeaJayAttribute.AccessType.class);
                    } else {
                        accessType = defaultAccessType;
                    }
                    if (typeValue != null) {
                        type = typeValue.getValue(TeaJayAttribute.Type.class);
                    } else {
                        type = defaultType;
                    }
                    TeaJayAttribute teaJayAttribute = new TeaJayAttribute(type, accessType);
                    if (info != null) {
                        System.out.println("Adding " + teaJayAttribute + " to " + info);
                        info.removeAttribute(TeaJayAttribute.NAME);
                        info.addAttribute(teaJayAttribute);
                    } else if (file != null) {
                        System.out.println("Adding " + teaJayAttribute + " to " + file.getThis().getClassName());
                        file.removeAttribute(TeaJayAttribute.NAME);
                        file.addAttribute(teaJayAttribute);
                    } else {
                        throw new Error("fatal error");
                    }
                }
            }

        }
    }

    private static void processSignatureOverride(RuntimeInvisibleAnnotations annotations, MethodInfo info, ClassFile file) throws ByteCodeException {
        if (annotations != null) {
            for (Annotation a : annotations) {
                if (a.getType().equals(overrideDescr)) {
                    ElementValue val = a.getElementValue("value");
                    if (val != null) {
                        String sig = val.getValue(String.class);
                        if (info != null) {
                            System.out.println(info + ": Overriding signature to " + info.getName().getValue() + " " + sig);
                            info.removeAttribute(Signature.NAME);
                            info.addAttribute(new Signature(sig));
                        } else if (file != null) {
                            System.out.println(file.getThis().getClassName() + ": Overriding class signature to " + sig);
                            file.removeAttribute(Signature.NAME);
                            file.addAttribute(new Signature(sig));
                        } else {
                            throw new Error("fatal error");
                        }
                    } else {
                        System.err.println("[WARN] Invalid annotation");
                    }
                }
            }
        }
    }

    public static void main(String[] args) {
        if (args.length <= 0) {
            System.err.println("Usage: java " + InjectTJAttribute.class.getName() + "[classfile or direcctory] ...");
            System.exit(1);
        }
        new FileIterator(args).scan(new FilesVisitor() {
            @Override
            public void visit(String javaPackage, String filename, Input in) throws StopScan {
                try {
                    inject(in.getFile());
                } catch (IOException | ByteCodeException ex) {
                    System.err.println(ex.getLocalizedMessage());
                    System.exit(1);
                }
            }

            @Override
            public void failed(File file, IOException reason) throws StopScan {
                System.err.println("[WARN] unable to read file " + file + ": " + reason);
                System.exit(1);
            }

            @Override
            public boolean accept(File file) {
                String name = file.getName().toLowerCase();
                return !file.isDirectory() && name.endsWith(".class");
            }
        }, false);
    }
}
