/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.util.tjattribute;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Objects;
import teajay.bytecode.ByteCodeWritable;
import teajay.bytecode.U4;
import teajay.bytecode.attributes.AttributeInfo;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.Utf8Info;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.exceptions.UTFOutOfBoundsException;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class TeaJayAttribute extends AttributeInfo {

    public static final String NAME = "TeaJay";
    private final Type type;
    private final AccessType atype;

    public enum Type implements ByteCodeWritable {
        Closure(6),
        Impl(5),
        Typedef(3),
        Class(4),
        Enum(1),
        Interface(2),
        TJBuiltinType(0),
        Hidden(-1);
        private final int data;
        
        private Type(int data) {
            this.data = data;
        }

        @Override
        public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
            out.writeByte(data);
        }

        public static Type byValue(byte val) throws ByteCodeException {
            for (Type t : Type.values()) {
                if (t.data == val) {
                    return t;
                }
            }
            throw new ByteCodeException("TeaJayAttribute: Unknown type: " + Integer.toHexString(val));
        }

        public int compare(Type t) {
            return Integer.compare(data, t.data);
        }
    }
    
    public enum AccessType implements ByteCodeWritable {
        None(0),
        Final(1),
        Abstract(2),
        Hidden(3);
        
        private final int data;
        
        private AccessType(int data) {
            this.data = data;
        }

        @Override
        public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
            out.writeByte(data);
        }

        public static AccessType byValue(byte val) throws ByteCodeException {
            for (AccessType t : AccessType.values()) {
                if (t.data == val) {
                    return t;
                }
            }
            throw new ByteCodeException("TeaJayAttribute: Unknown access type: " + Integer.toHexString(val));
        }
    }
    
    public TeaJayAttribute(Type type, AccessType atype) throws UTFOutOfBoundsException {
        this(new Utf8Info(NAME), type, atype);
    }

    public TeaJayAttribute(Type type) throws UTFOutOfBoundsException {
        this(new Utf8Info(NAME), type, AccessType.None);
    }

    private TeaJayAttribute(Utf8Info name, Type type, AccessType atype) throws UTFOutOfBoundsException {
        super(name);
        checkName(NAME, name);
        this.type = type;
        this.atype = atype;
    }

    @Override
    public U4 getLength(ConstantPool cpool) throws ByteCodeException {
        return new U4(2);
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        super.write(out, cpool);
        type.write(out, cpool);
        atype.write(out, cpool);
    }

    public Type getType() {
        return type;
    }
    
    public AccessType getAccessType() {
        return atype;
    }

    @Override
    public String toString() {
        return "TeaJayAttribute: " + type + " ( " + atype + " )";
    }
    
    @Override
    public boolean equals(Object obj) {
        if (obj instanceof TeaJayAttribute) {
            return ((TeaJayAttribute) obj).type == type && ((TeaJayAttribute) obj).atype == atype;
        }
        return false;
    }

    @Override
    public int hashCode() {
        int hash = 7;
        hash = 29 * hash + Objects.hashCode(this.type);
        hash = 29 * hash + Objects.hashCode(this.atype);
        return hash;
    }

    public static TeaJayAttribute read(DataInputStream in, ConstantPool cpool, int major, int minor, int length, Utf8Info name) throws ByteCodeException, IOException {
        return new TeaJayAttribute(name, Type.byValue(in.readByte()), AccessType.byValue(in.readByte()));
    }
}
